import logging
import os
import shutil
import time
import uuid
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence

import lightning.pytorch as pl
import torch
from lightning.pytorch.callbacks import BasePredictionWriter

from ..constants import (
    AUG_LOGITS,
    BBOX,
    LOGIT_SCALE,
    MULTIMODAL_FEATURES,
    MULTIMODAL_FEATURES_POST_AUG,
    MULTIMODAL_FEATURES_PRE_AUG,
    ORI_LOGITS,
    VAE_MEAN,
    VAE_VAR,
    WEIGHT,
)

logger = logging.getLogger(__name__)


class DDPPredictionWriter(BasePredictionWriter):
    def __init__(
        self, output_dir: Optional[str], write_interval: Optional[str], strategy: Optional[str], sleep_time=5
    ):
        """
        Parameters
        ----------
        output_dir
            The directory to save predictions.
        write_interval
            When to write. Could be "batch" or "epoch".
            See Lightning's source code at
            https://lightning.ai/docs/pytorch/stable/api/lightning.pytorch.callbacks.BasePredictionWriter.html
        sleep_time
            If other process does not finish writing, sleep for a few seconds and recheck.
        """
        super().__init__(write_interval)
        if output_dir is None:  # TODO: give the predictor a default save_path before calling DDPPredictionWriter
            output_dir = "temp_cache_for_ddp_prediction"
            logging.warning(
                f"Current predictor's save_path is None, using a default cache folder which may cause an error in prediction I/O. Try init the predictor with a save_path."
            )
        assert isinstance(
            output_dir, (str, Path)
        ), f"Only str and pathlib.Path types are supported for path, but got {output_dir} of type {type(output_dir)}."
        self.sleep_time = sleep_time
        output_dir = os.path.abspath(os.path.expanduser(output_dir))
        if "spawn" in strategy:
            self.output_dir = os.path.join(output_dir, f"predictions_{uuid.uuid4().hex}")
        else:
            self.output_dir = os.path.join(output_dir, "ddp_prediction_cache")
        try:
            os.makedirs(self.output_dir, exist_ok=False)
        except FileExistsError:  # assume the string is unique
            logger.warning(
                f"{self.output_dir} already exists. This could be caused by DDP subprocess. Just make sure the previous cache is removed in {self.output_dir}."
            )

    def get_predictions_cache_dir(self, global_rank: int):
        """
        Parameters
        ----------
        global_rank
            Global rank of current process.
        """
        return os.path.join(self.output_dir, f"predictions_rank_{global_rank}.pt")

    def get_batch_indices_cache_dir(self, global_rank: int):
        """
        Parameters
        ----------
        global_rank
            Global rank of current process.
        """
        return os.path.join(self.output_dir, f"sample_indices_rank_{global_rank}.pt")

    def write_on_epoch_end(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        predictions: Sequence[Any],
        batch_indices: Sequence[Any],
    ):
        """
        Parameters
        ----------
        trainer
            Pytorch Lightning trainer.
        pl_module
            Pytorch Lightning module.
        predictions
            The predicted results.
        batch_indices
            The corresponding batch indices for prediction results.
        """
        # this will create N (num processes) files in `cache_dir` each containing
        # the predictions of its respective rank
        torch.save(predictions, self.get_predictions_cache_dir(trainer.global_rank))  # nosec B614
        # here we save `batch_indices` to get the information about the data index
        # from prediction data
        torch.save(batch_indices, self.get_batch_indices_cache_dir(trainer.global_rank))  # nosec B614

    def read_single_gpu_results(self, global_rank: Optional[int]):
        """
        Parameters
        ----------
        global_rank
            Global rank of current process.
        """
        sample_indices_file = self.get_batch_indices_cache_dir(global_rank)
        predictions_file = self.get_predictions_cache_dir(global_rank)
        while (not os.path.exists(sample_indices_file)) or (not os.path.exists(predictions_file)):
            logger.info(f"waiting for rank #{global_rank} to finish saving predictions...")
            time.sleep(self.sleep_time)
        sample_indices = torch.load(sample_indices_file)  # nosec B614
        predictions = torch.load(predictions_file)  # nosec B614

        return sample_indices, predictions

    def flatten(self, x):
        """
        Flatten nested lists into one list.

        Parameters
        ----------
        x
            A nested list to be flattened.
        """
        if isinstance(x, list):
            return [a for i in x for a in self.flatten(i)]
        else:
            return [x]

    def collate(self, x: List[Dict]):
        """
        Collate a list of dictionaries into one.
        Each value is a tensor concatenated from a list of batch tensors.

        Parameters
        ----------
        x
            A list of batch results. Each batch is one (nested) dictionary.
        """
        if BBOX in x[0]:  # for detection outputs just sort the batches
            return x

        results = dict()
        if len(x[0]) == 0:  # dict is empty
            return dict()

        for k, v in x[0].items():
            if k in [
                WEIGHT,
                LOGIT_SCALE,
                MULTIMODAL_FEATURES,
                MULTIMODAL_FEATURES_PRE_AUG,
                MULTIMODAL_FEATURES_POST_AUG,
                ORI_LOGITS,
                AUG_LOGITS,
                VAE_MEAN,
                VAE_VAR,
            ]:  # ignore the keys
                continue
            elif isinstance(v, dict):
                results[k] = self.collate([i[k] for i in x])
            elif isinstance(v, torch.Tensor):
                results[k] = torch.cat([i[k] for i in x])
            else:
                raise ValueError(
                    f"Unsupported data type {type(v)} is found in prediction results. "
                    f"We only support dictionary with torch tensor values."
                )

        return results

    def sort(self, x: Dict, indices: List):
        """
        Sort the values of each key according to the indices.
        This is to make sure that prediction results follow the order of input samples.

        Parameters
        ----------
        x
            A dictionary with all the predictions.
        indices
            A list of indices.
        """
        if isinstance(x, list) and BBOX in x[0]:  # for detection outputs just sort the batches
            return [xi for _, xi in sorted(zip(indices, x), key=lambda ele: ele[0])]

        results = dict()
        for k, v in x.items():
            if isinstance(v, dict):
                results[k] = self.sort(v, indices)
            else:
                assert len(indices) == len(
                    v
                ), f"Size mismatch, {k}: {v} of len {len(v)} and indices {indices} of length {len(indices)}"
                results[k] = [x for _, x in sorted(zip(indices, v), key=lambda ele: ele[0])]
                results[k] = torch.stack(results[k])

        return results

    def collect_all_gpu_results(self, num_gpus):
        """
        Parameters
        ----------
        num_gpus
            Number of gpus used.
        """
        sample_indices = []
        predictions = []
        for global_rank in range(num_gpus):
            sample_indices_per_rank, predictions_per_rank = self.read_single_gpu_results(global_rank)
            sample_indices += self.flatten(sample_indices_per_rank)
            predictions += self.flatten(predictions_per_rank)

        assert sorted(sample_indices) == list(range(len(sample_indices)))
        predictions = self.collate(predictions)
        sorted_predictions = self.sort(predictions, indices=sample_indices)
        shutil.rmtree(self.output_dir)

        return [sorted_predictions]



_base_ = "./yolox_s_8xb8-300e_coco.py"

# model settings
model = dict(
    data_preprocessor=dict(
        batch_augments=[dict(type="BatchSyncRandomResize", random_size_range=(320, 640), size_divisor=32, interval=10)]
    ),
    backbone=dict(deepen_factor=0.33, widen_factor=0.375),
    neck=dict(in_channels=[96, 192, 384], out_channels=96),
    bbox_head=dict(in_channels=96, feat_channels=96),
)

img_scale = (640, 640)  # width, height

test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="Resize", scale=(416, 416), keep_ratio=True),
    dict(type="Pad", pad_to_square=True, pad_val=dict(img=(114.0, 114.0, 114.0))),
    dict(type="LoadAnnotations", with_bbox=True),
    dict(type="PackDetInputs", meta_keys=("img_id", "img_path", "ori_shape", "img_shape", "scale_factor")),
]



import collections
import copy
import logging
import math
from typing import Dict, List, Optional, Sequence, Tuple, Union

import mmcv
import numpy as np
import pandas as pd
import torch
from mmcv.transforms import BaseTransform
from mmcv.transforms.utils import cache_randomness
from mmdet.structures.bbox import autocast_box_type
from mmdet.utils import log_img_scale
from mmengine.config import Config as MMConfig
from mmengine.dataset import BaseDataset
from numpy import random

from ...constants import GET_ITEM_ERROR_RETRY, MULTI_IMAGE_MIX_DATASET, ROIS
from ..preprocess_dataframe import MultiModalFeaturePreprocessor
from ..utils import apply_data_processor, apply_df_preprocessor, get_per_sample_features

logger = logging.getLogger(__name__)


class MultiImageMixDataset(torch.utils.data.Dataset):
    """
    A Pytorch DataSet class to process a multimodal pd.DataFrame. It first uses a preprocessor to
    produce model-agnostic features. Then, each processor prepares customized data for one modality
    per model. For code simplicity, here we treat ground-truth label as one modality. This class is
    independent of specific data modalities and models.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        preprocessor: List[MultiModalFeaturePreprocessor],
        processors: List[dict],
        model_config: MMConfig,
        id_mappings: Optional[Union[Dict[str, Dict], Dict[str, pd.Series]]] = None,
        is_training: bool = False,
    ):
        """
        Parameters
        ----------
        data
            A pd.DataFrame containing multimodal features.
        preprocessor
            A list of multimodal feature preprocessors generating model-agnostic features.
        processors
            Data processors customizing data for each modality per model.
        id_mappings
            Id-to-content mappings. The contents can be text, image, etc.
            This is used when the dataframe contains the query/response indexes instead of their contents.
        is_training
            Whether in training mode. Some data processing may be different between training
            and validation/testing/prediction, e.g., image data augmentation is used only in
            training.
        model_config
            Model config used to decided dataset type. e.g. if multi_image_mix_dataset is used in detection model,
            MultiImageMixDataset will be used instead of BaseDataset
        """
        super().__init__()
        self.processors = processors
        self.is_training = is_training
        self._consecutive_errors = 0

        mix_config = model_config[MULTI_IMAGE_MIX_DATASET]
        self.mix_data_key = "mmdet_image_image"  # the key of the data to mix, TODO: remove hardcoding
        self.mix_result_key = "mix_results"  # the key of the mix result to store

        self.mix_transforms = []
        self.mix_transforms_types = []  # TODO: remove hardcode
        if "mosaic" in mix_config:
            self.mix_transforms.append(Mosaic(**mix_config["mosaic"]))
            self.mix_transforms_types.append("mosaic")
        if "mixup" in mix_config:
            self.mix_transforms.append(MixUp(**mix_config["mixup"]))
            self.mix_transforms_types.append("mixup")

        self._skip_type_keys = None  # TODO: remove hardcode, we need to disable multi image mix in late epochs
        self.max_refetch = 15  # TODO: remove hardcode (do we need refetch?)

        self.lengths = []

        for i, (per_preprocessor, per_processors_group) in enumerate(zip(preprocessor, processors)):
            modality_features, modality_types, length = apply_df_preprocessor(
                data=data,
                df_preprocessor=per_preprocessor,
                modalities=per_processors_group.keys(),
            )
            self.lengths.append(length)
            setattr(self, f"modality_features_{i}", modality_features)
            setattr(self, f"modality_types_{i}", modality_types)

        assert len(set(self.lengths)) == 1

        self.id_mappings = id_mappings

    def __len__(self):
        """
        Assume that all modalities have the same sample number.

        Returns
        -------
        Sample number in this dataset.
        """
        return self.lengths[0]

    def _load_item(self, idx):
        """
        Get a single item without mix_results.
        Iterate through all data processors to prepare model inputs. The data processors are
        organized first by modalities and then by models.

        Parameters
        ----------
        idx
            Index of sample to process.

        Returns
        -------
        Input data formatted as a dictionary.
        """
        ret = dict()
        try:
            for group_id, per_processors_group in enumerate(self.processors):
                per_sample_features = get_per_sample_features(
                    modality_features=getattr(self, f"modality_features_{group_id}"),
                    modality_types=getattr(self, f"modality_types_{group_id}"),
                    idx=idx,
                    id_mappings=self.id_mappings,
                )
                per_ret = apply_data_processor(
                    per_sample_features=per_sample_features,
                    data_processors=per_processors_group,
                    feature_modalities=getattr(self, f"modality_types_{group_id}"),
                    is_training=self.is_training,
                    load_only=True,
                )
                ret.update(per_ret)
        except Exception as e:
            logger.debug(f"Skipping sample {idx} due to '{e}'")
            self._consecutive_errors += 1
            if self._consecutive_errors < GET_ITEM_ERROR_RETRY:
                return self.__getitem__((idx + 1) % self.__len__())
            else:
                raise e
        self._consecutive_errors = 0

        return ret

    def __getitem__(self, idx):
        """
        Iterate through all data processors to prepare model inputs. The data processors are
        organized first by modalities and then by models.

        Parameters
        ----------
        idx
            Index of sample to process.

        Returns
        -------
        Input data formatted as a dictionary.
        """
        results = copy.deepcopy(self._load_item(idx))

        for transform, transform_type in zip(self.mix_transforms, self.mix_transforms_types):
            assert hasattr(transform, "get_indexes")

            if self._skip_type_keys is not None and transform_type in self._skip_type_keys:
                continue

            for i in range(self.max_refetch):
                # Make sure the results passed the loading pipeline
                # of the original dataset is not None.
                indexes = transform.get_indexes(self)
                if not isinstance(indexes, collections.abc.Sequence):
                    indexes = [indexes]
                mix_results = [copy.deepcopy(self._load_item(index)[self.mix_data_key]) for index in indexes]
                if None not in mix_results:
                    results[self.mix_data_key][self.mix_result_key] = mix_results
                    break
            else:
                raise RuntimeError(
                    "The loading pipeline of the original dataset"
                    " always return None. Please check the correctness "
                    "of the dataset and its pipeline."
                )

            for i in range(self.max_refetch):
                # To confirm the results passed the training pipeline
                # of the wrapper is not None.
                updated_results = transform(copy.deepcopy(results[self.mix_data_key]))
                if updated_results is not None:
                    results[self.mix_data_key] = updated_results
                    break
            else:
                raise RuntimeError(
                    "The training pipeline of the dataset wrapper"
                    " always return None.Please check the correctness "
                    "of the dataset and its pipeline."
                )

            if self.mix_result_key in results[self.mix_data_key]:
                results[self.mix_data_key].pop(self.mix_result_key)

        rois_processor = self.processors[0][ROIS][0]  # TODO: remove hardcode
        results.update(
            rois_processor.process_one_loaded_sample(
                results,
                is_training=True,  # This dataset is used only in training
            )
        )

        return results


class Mosaic(BaseTransform):
    """Mosaic augmentation.

    Given 4 images, mosaic transform combines them into
    one output image. The output image is composed of the parts from each sub-
    image.

    .. code:: text

                        mosaic transform
                        center_x
                +------------------------------+
                |       pad        |  pad      |
                |      +-----------+           |
                |      |           |           |
                |      |  image1   |--------+  |
                |      |           |        |  |
                |      |           | image2 |  |
    center_y   |----+-------------+-----------|
                |    |   cropped   |           |
                |pad |   image3    |  image4   |
                |    |             |           |
                +----|-------------+-----------+
                    |             |
                    +-------------+

    The mosaic transform steps are as follows:

        1. Choose the mosaic center as the intersections of 4 images
        2. Get the left top image according to the index, and randomly
            sample another 3 images from the custom dataset.
        3. Sub image will be cropped if image is larger than mosaic patch

    Required Keys:

    - img
    - gt_bboxes (BaseBoxes[torch.float32]) (optional)
    - gt_bboxes_labels (np.int64) (optional)
    - gt_ignore_flags (bool) (optional)
    - mix_results (List[dict])

    Modified Keys:

    - img
    - img_shape
    - gt_bboxes (optional)
    - gt_bboxes_labels (optional)
    - gt_ignore_flags (optional)

    Args:
        img_scale (Sequence[int]): Image size after mosaic pipeline of single
            image. The shape order should be (width, height).
            Defaults to (640, 640).
        center_ratio_range (Sequence[float]): Center ratio range of mosaic
            output. Defaults to (0.5, 1.5).
        bbox_clip_border (bool, optional): Whether to clip the objects outside
            the border of the image. In some dataset like MOT17, the gt bboxes
            are allowed to cross the border of images. Therefore, we don't
            need to clip the gt bboxes in these cases. Defaults to True.
        pad_val (int): Pad value. Defaults to 114.
        prob (float): Probability of applying this transformation.
            Defaults to 1.0.
    """

    def __init__(
        self,
        img_scale: Tuple[int, int] = (640, 640),
        center_ratio_range: Tuple[float, float] = (0.5, 1.5),
        bbox_clip_border: bool = True,
        pad_val: float = 114.0,
        prob: float = 1.0,
    ) -> None:
        assert isinstance(img_scale, tuple)
        assert 0 <= prob <= 1.0, "The probability should be in range [0,1]. " f"got {prob}."

        log_img_scale(img_scale, skip_square=True, shape_order="wh")
        self.img_scale = img_scale
        self.center_ratio_range = center_ratio_range
        self.bbox_clip_border = bbox_clip_border
        self.pad_val = pad_val
        self.prob = prob

    @cache_randomness
    def get_indexes(self, dataset: BaseDataset) -> int:
        """Call function to collect indexes.

        Args:
            dataset (:obj:`MultiImageMixDataset`): The dataset.

        Returns:
            list: indexes.
        """

        indexes = [random.randint(0, len(dataset)) for _ in range(3)]
        return indexes

    @autocast_box_type()
    def transform(self, results: dict) -> dict:
        """Mosaic transform function.

        Args:
            results (dict): Result dict.

        Returns:
            dict: Updated result dict.
        """
        if random.uniform(0, 1) > self.prob:
            return results

        assert "mix_results" in results
        mosaic_bboxes = []
        mosaic_bboxes_labels = []
        mosaic_ignore_flags = []
        if len(results["img"].shape) == 3:
            mosaic_img = np.full(
                (int(self.img_scale[1] * 2), int(self.img_scale[0] * 2), 3),
                self.pad_val,
                dtype=results["img"].dtype,
            )
        else:
            mosaic_img = np.full(
                (int(self.img_scale[1] * 2), int(self.img_scale[0] * 2)), self.pad_val, dtype=results["img"].dtype
            )

        # mosaic center x, y
        center_x = int(random.uniform(*self.center_ratio_range) * self.img_scale[0])
        center_y = int(random.uniform(*self.center_ratio_range) * self.img_scale[1])
        center_position = (center_x, center_y)

        loc_strs = ("top_left", "top_right", "bottom_left", "bottom_right")
        for i, loc in enumerate(loc_strs):
            if loc == "top_left":
                results_patch = copy.deepcopy(results)
            else:
                results_patch = copy.deepcopy(results["mix_results"][i - 1])

            img_i = results_patch["img"]
            h_i, w_i = img_i.shape[:2]
            # keep_ratio resize
            scale_ratio_i = min(self.img_scale[1] / h_i, self.img_scale[0] / w_i)
            img_i = mmcv.imresize(img_i, (int(w_i * scale_ratio_i), int(h_i * scale_ratio_i)))

            # compute the combine parameters
            paste_coord, crop_coord = self._mosaic_combine(loc, center_position, img_i.shape[:2][::-1])
            x1_p, y1_p, x2_p, y2_p = paste_coord
            x1_c, y1_c, x2_c, y2_c = crop_coord

            # crop and paste image
            mosaic_img[y1_p:y2_p, x1_p:x2_p] = img_i[y1_c:y2_c, x1_c:x2_c]

            # adjust coordinate
            gt_bboxes_i = results_patch["gt_bboxes"]
            gt_bboxes_labels_i = results_patch["gt_bboxes_labels"]
            gt_ignore_flags_i = results_patch["gt_ignore_flags"]

            padw = x1_p - x1_c
            padh = y1_p - y1_c
            gt_bboxes_i.rescale_([scale_ratio_i, scale_ratio_i])
            gt_bboxes_i.translate_([padw, padh])
            mosaic_bboxes.append(gt_bboxes_i)
            mosaic_bboxes_labels.append(gt_bboxes_labels_i)
            mosaic_ignore_flags.append(gt_ignore_flags_i)

        mosaic_bboxes = mosaic_bboxes[0].cat(mosaic_bboxes, 0)
        mosaic_bboxes_labels = np.concatenate(mosaic_bboxes_labels, 0)
        mosaic_ignore_flags = np.concatenate(mosaic_ignore_flags, 0)

        if self.bbox_clip_border:
            mosaic_bboxes.clip_([2 * self.img_scale[1], 2 * self.img_scale[0]])
        # remove outside bboxes
        inside_inds = mosaic_bboxes.is_inside([2 * self.img_scale[1], 2 * self.img_scale[0]]).numpy()
        mosaic_bboxes = mosaic_bboxes[inside_inds]
        mosaic_bboxes_labels = mosaic_bboxes_labels[inside_inds]
        mosaic_ignore_flags = mosaic_ignore_flags[inside_inds]

        results["img"] = mosaic_img
        results["img_shape"] = mosaic_img.shape[:2]
        results["gt_bboxes"] = mosaic_bboxes
        results["gt_bboxes_labels"] = mosaic_bboxes_labels
        results["gt_ignore_flags"] = mosaic_ignore_flags
        return results

    def _mosaic_combine(
        self, loc: str, center_position_xy: Sequence[float], img_shape_wh: Sequence[int]
    ) -> Tuple[Tuple[int], Tuple[int]]:
        """Calculate global coordinate of mosaic image and local coordinate of
        cropped sub-image.

        Args:
            loc (str): Index for the sub-image, loc in ('top_left',
            'top_right', 'bottom_left', 'bottom_right').
            center_position_xy (Sequence[float]): Mixing center for 4 images,
                (x, y).
            img_shape_wh (Sequence[int]): Width and height of sub-image

        Returns:
            tuple[tuple[float]]: Corresponding coordinate of pasting and
                cropping
                - paste_coord (tuple): paste corner coordinate in mosaic image.
                - crop_coord (tuple): crop corner coordinate in mosaic image.
        """
        assert loc in ("top_left", "top_right", "bottom_left", "bottom_right")
        if loc == "top_left":
            # index0 to top left part of image
            x1, y1, x2, y2 = (
                max(center_position_xy[0] - img_shape_wh[0], 0),
                max(center_position_xy[1] - img_shape_wh[1], 0),
                center_position_xy[0],
                center_position_xy[1],
            )
            crop_coord = img_shape_wh[0] - (x2 - x1), img_shape_wh[1] - (y2 - y1), img_shape_wh[0], img_shape_wh[1]

        elif loc == "top_right":
            # index1 to top right part of image
            x1, y1, x2, y2 = (
                center_position_xy[0],
                max(center_position_xy[1] - img_shape_wh[1], 0),
                min(center_position_xy[0] + img_shape_wh[0], self.img_scale[0] * 2),
                center_position_xy[1],
            )
            crop_coord = 0, img_shape_wh[1] - (y2 - y1), min(img_shape_wh[0], x2 - x1), img_shape_wh[1]

        elif loc == "bottom_left":
            # index2 to bottom left part of image
            x1, y1, x2, y2 = (
                max(center_position_xy[0] - img_shape_wh[0], 0),
                center_position_xy[1],
                center_position_xy[0],
                min(self.img_scale[1] * 2, center_position_xy[1] + img_shape_wh[1]),
            )
            crop_coord = img_shape_wh[0] - (x2 - x1), 0, img_shape_wh[0], min(y2 - y1, img_shape_wh[1])

        else:
            # index3 to bottom right part of image
            x1, y1, x2, y2 = (
                center_position_xy[0],
                center_position_xy[1],
                min(center_position_xy[0] + img_shape_wh[0], self.img_scale[0] * 2),
                min(self.img_scale[1] * 2, center_position_xy[1] + img_shape_wh[1]),
            )
            crop_coord = 0, 0, min(img_shape_wh[0], x2 - x1), min(y2 - y1, img_shape_wh[1])

        paste_coord = x1, y1, x2, y2
        return paste_coord, crop_coord

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f"(img_scale={self.img_scale}, "
        repr_str += f"center_ratio_range={self.center_ratio_range}, "
        repr_str += f"pad_val={self.pad_val}, "
        repr_str += f"prob={self.prob})"
        return repr_str


class MixUp(BaseTransform):
    """MixUp data augmentation.

    .. code:: text

                        mixup transform
                +------------------------------+
                | mixup image   |              |
                |      +--------|--------+     |
                |      |        |        |     |
                |---------------+        |     |
                |      |                 |     |
                |      |      image      |     |
                |      |                 |     |
                |      |                 |     |
                |      |-----------------+     |
                |             pad              |
                +------------------------------+

    The mixup transform steps are as follows:

        1. Another random image is picked by dataset and embedded in
        the top left patch(after padding and resizing)
        2. The target of mixup transform is the weighted average of mixup
        image and origin image.

    Required Keys:

    - img
    - gt_bboxes (BaseBoxes[torch.float32]) (optional)
    - gt_bboxes_labels (np.int64) (optional)
    - gt_ignore_flags (bool) (optional)
    - mix_results (List[dict])


    Modified Keys:

    - img
    - img_shape
    - gt_bboxes (optional)
    - gt_bboxes_labels (optional)
    - gt_ignore_flags (optional)


    Args:
        img_scale (Sequence[int]): Image output size after mixup pipeline.
            The shape order should be (width, height). Defaults to (640, 640).
        ratio_range (Sequence[float]): Scale ratio of mixup image.
            Defaults to (0.5, 1.5).
        flip_ratio (float): Horizontal flip ratio of mixup image.
            Defaults to 0.5.
        pad_val (int): Pad value. Defaults to 114.
        max_iters (int): The maximum number of iterations. If the number of
            iterations is greater than `max_iters`, but gt_bbox is still
            empty, then the iteration is terminated. Defaults to 15.
        bbox_clip_border (bool, optional): Whether to clip the objects outside
            the border of the image. In some dataset like MOT17, the gt bboxes
            are allowed to cross the border of images. Therefore, we don't
            need to clip the gt bboxes in these cases. Defaults to True.
    """

    def __init__(
        self,
        img_scale: Tuple[int, int] = (640, 640),
        ratio_range: Tuple[float, float] = (0.5, 1.5),
        flip_ratio: float = 0.5,
        pad_val: float = 114.0,
        max_iters: int = 15,
        bbox_clip_border: bool = True,
    ) -> None:
        assert isinstance(img_scale, tuple)
        log_img_scale(img_scale, skip_square=True, shape_order="wh")
        self.dynamic_scale = img_scale
        self.ratio_range = ratio_range
        self.flip_ratio = flip_ratio
        self.pad_val = pad_val
        self.max_iters = max_iters
        self.bbox_clip_border = bbox_clip_border

    @cache_randomness
    def get_indexes(self, dataset: BaseDataset) -> int:
        """Call function to collect indexes.

        Args:
            dataset (:obj:`MultiImageMixDataset`): The dataset.

        Returns:
            list: indexes.
        """

        index = [np.random.randint(0, len(dataset)) for _ in range(1)]

        return index

    @autocast_box_type()
    def transform(self, results: dict) -> dict:
        """MixUp transform function.

        Args:
            results (dict): Result dict.

        Returns:
            dict: Updated result dict.
        """

        assert "mix_results" in results
        assert len(results["mix_results"]) == 1, "MixUp only support 2 images now !"

        if results["mix_results"][0]["gt_bboxes"].shape[0] == 0:
            # empty bbox
            return results

        retrieve_results = results["mix_results"][0]
        retrieve_img = retrieve_results["img"]

        jit_factor = random.uniform(*self.ratio_range)
        is_filp = random.uniform(0, 1) > self.flip_ratio

        if len(retrieve_img.shape) == 3:
            out_img = (
                np.ones((self.dynamic_scale[1], self.dynamic_scale[0], 3), dtype=retrieve_img.dtype) * self.pad_val
            )
        else:
            out_img = np.ones(self.dynamic_scale[::-1], dtype=retrieve_img.dtype) * self.pad_val

        # 1. keep_ratio resize
        scale_ratio = min(self.dynamic_scale[1] / retrieve_img.shape[0], self.dynamic_scale[0] / retrieve_img.shape[1])
        retrieve_img = mmcv.imresize(
            retrieve_img, (int(retrieve_img.shape[1] * scale_ratio), int(retrieve_img.shape[0] * scale_ratio))
        )

        # 2. paste
        out_img[: retrieve_img.shape[0], : retrieve_img.shape[1]] = retrieve_img

        # 3. scale jit
        scale_ratio *= jit_factor
        out_img = mmcv.imresize(out_img, (int(out_img.shape[1] * jit_factor), int(out_img.shape[0] * jit_factor)))

        # 4. flip
        if is_filp:
            out_img = out_img[:, ::-1, :]

        # 5. random crop
        ori_img = results["img"]
        origin_h, origin_w = out_img.shape[:2]
        target_h, target_w = ori_img.shape[:2]
        padded_img = np.ones((max(origin_h, target_h), max(origin_w, target_w), 3)) * self.pad_val
        padded_img = padded_img.astype(np.uint8)
        padded_img[:origin_h, :origin_w] = out_img

        x_offset, y_offset = 0, 0
        if padded_img.shape[0] > target_h:
            y_offset = random.randint(0, padded_img.shape[0] - target_h)
        if padded_img.shape[1] > target_w:
            x_offset = random.randint(0, padded_img.shape[1] - target_w)
        padded_cropped_img = padded_img[y_offset : y_offset + target_h, x_offset : x_offset + target_w]

        # 6. adjust bbox
        retrieve_gt_bboxes = retrieve_results["gt_bboxes"]
        retrieve_gt_bboxes.rescale_([scale_ratio, scale_ratio])
        if self.bbox_clip_border:
            retrieve_gt_bboxes.clip_([origin_h, origin_w])

        if is_filp:
            retrieve_gt_bboxes.flip_([origin_h, origin_w], direction="horizontal")

        # 7. filter
        cp_retrieve_gt_bboxes = retrieve_gt_bboxes.clone()
        cp_retrieve_gt_bboxes.translate_([-x_offset, -y_offset])
        if self.bbox_clip_border:
            cp_retrieve_gt_bboxes.clip_([target_h, target_w])

        # 8. mix up
        ori_img = ori_img.astype(np.float32)
        mixup_img = 0.5 * ori_img + 0.5 * padded_cropped_img.astype(np.float32)

        retrieve_gt_bboxes_labels = retrieve_results["gt_bboxes_labels"]
        retrieve_gt_ignore_flags = retrieve_results["gt_ignore_flags"]

        mixup_gt_bboxes = cp_retrieve_gt_bboxes.cat((results["gt_bboxes"], cp_retrieve_gt_bboxes), dim=0)
        mixup_gt_bboxes_labels = np.concatenate((results["gt_bboxes_labels"], retrieve_gt_bboxes_labels), axis=0)
        mixup_gt_ignore_flags = np.concatenate((results["gt_ignore_flags"], retrieve_gt_ignore_flags), axis=0)

        # remove outside bbox
        inside_inds = mixup_gt_bboxes.is_inside([target_h, target_w]).numpy()
        mixup_gt_bboxes = mixup_gt_bboxes[inside_inds]
        mixup_gt_bboxes_labels = mixup_gt_bboxes_labels[inside_inds]
        mixup_gt_ignore_flags = mixup_gt_ignore_flags[inside_inds]

        results["img"] = mixup_img.astype(np.uint8)
        results["img_shape"] = mixup_img.shape[:2]
        results["gt_bboxes"] = mixup_gt_bboxes
        results["gt_bboxes_labels"] = mixup_gt_bboxes_labels
        results["gt_ignore_flags"] = mixup_gt_ignore_flags

        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f"(dynamic_scale={self.dynamic_scale}, "
        repr_str += f"ratio_range={self.ratio_range}, "
        repr_str += f"flip_ratio={self.flip_ratio}, "
        repr_str += f"pad_val={self.pad_val}, "
        repr_str += f"max_iters={self.max_iters}, "
        repr_str += f"bbox_clip_border={self.bbox_clip_border})"
        return repr_str


class RandomAffine(BaseTransform):
    """Random affine transform data augmentation.

    This operation randomly generates affine transform matrix which including
    rotation, translation, shear and scaling transforms.

    Required Keys:

    - img
    - gt_bboxes (BaseBoxes[torch.float32]) (optional)
    - gt_bboxes_labels (np.int64) (optional)
    - gt_ignore_flags (bool) (optional)

    Modified Keys:

    - img
    - img_shape
    - gt_bboxes (optional)
    - gt_bboxes_labels (optional)
    - gt_ignore_flags (optional)

    Args:
        max_rotate_degree (float): Maximum degrees of rotation transform.
            Defaults to 10.
        max_translate_ratio (float): Maximum ratio of translation.
            Defaults to 0.1.
        scaling_ratio_range (tuple[float]): Min and max ratio of
            scaling transform. Defaults to (0.5, 1.5).
        max_shear_degree (float): Maximum degrees of shear
            transform. Defaults to 2.
        border (tuple[int]): Distance from width and height sides of input
            image to adjust output shape. Only used in mosaic dataset.
            Defaults to (0, 0).
        border_val (tuple[int]): Border padding values of 3 channels.
            Defaults to (114, 114, 114).
        bbox_clip_border (bool, optional): Whether to clip the objects outside
            the border of the image. In some dataset like MOT17, the gt bboxes
            are allowed to cross the border of images. Therefore, we don't
            need to clip the gt bboxes in these cases. Defaults to True.
    """

    def __init__(
        self,
        max_rotate_degree: float = 10.0,
        max_translate_ratio: float = 0.1,
        scaling_ratio_range: Tuple[float, float] = (0.5, 1.5),
        max_shear_degree: float = 2.0,
        border: Tuple[int, int] = (0, 0),
        border_val: Tuple[int, int, int] = (114, 114, 114),
        bbox_clip_border: bool = True,
    ) -> None:
        assert 0 <= max_translate_ratio <= 1
        assert scaling_ratio_range[0] <= scaling_ratio_range[1]
        assert scaling_ratio_range[0] > 0
        self.max_rotate_degree = max_rotate_degree
        self.max_translate_ratio = max_translate_ratio
        self.scaling_ratio_range = scaling_ratio_range
        self.max_shear_degree = max_shear_degree
        self.border = border
        self.border_val = border_val
        self.bbox_clip_border = bbox_clip_border

    @cache_randomness
    def _get_random_homography_matrix(self, height, width):
        # Rotation
        rotation_degree = random.uniform(-self.max_rotate_degree, self.max_rotate_degree)
        rotation_matrix = self._get_rotation_matrix(rotation_degree)

        # Scaling
        scaling_ratio = random.uniform(self.scaling_ratio_range[0], self.scaling_ratio_range[1])
        scaling_matrix = self._get_scaling_matrix(scaling_ratio)

        # Shear
        x_degree = random.uniform(-self.max_shear_degree, self.max_shear_degree)
        y_degree = random.uniform(-self.max_shear_degree, self.max_shear_degree)
        shear_matrix = self._get_shear_matrix(x_degree, y_degree)

        # Translation
        trans_x = random.uniform(-self.max_translate_ratio, self.max_translate_ratio) * width
        trans_y = random.uniform(-self.max_translate_ratio, self.max_translate_ratio) * height
        translate_matrix = self._get_translation_matrix(trans_x, trans_y)

        warp_matrix = translate_matrix @ shear_matrix @ rotation_matrix @ scaling_matrix
        return warp_matrix

    @autocast_box_type()
    def transform(self, results: dict) -> dict:
        import cv2  # TODO: support random affine requires cv2

        img = results["img"]
        height = img.shape[0] + self.border[1] * 2
        width = img.shape[1] + self.border[0] * 2

        warp_matrix = self._get_random_homography_matrix(height, width)

        img = cv2.warpPerspective(img, warp_matrix, dsize=(width, height), borderValue=self.border_val)
        results["img"] = img
        results["img_shape"] = img.shape[:2]

        bboxes = results["gt_bboxes"]
        num_bboxes = len(bboxes)
        if num_bboxes:
            bboxes.project_(warp_matrix)
            if self.bbox_clip_border:
                bboxes.clip_([height, width])
            # remove outside bbox
            valid_index = bboxes.is_inside([height, width]).numpy()
            results["gt_bboxes"] = bboxes[valid_index]
            results["gt_bboxes_labels"] = results["gt_bboxes_labels"][valid_index]
            results["gt_ignore_flags"] = results["gt_ignore_flags"][valid_index]

            if "gt_masks" in results:
                raise NotImplementedError("RandomAffine only supports bbox.")
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f"(max_rotate_degree={self.max_rotate_degree}, "
        repr_str += f"max_translate_ratio={self.max_translate_ratio}, "
        repr_str += f"scaling_ratio_range={self.scaling_ratio_range}, "
        repr_str += f"max_shear_degree={self.max_shear_degree}, "
        repr_str += f"border={self.border}, "
        repr_str += f"border_val={self.border_val}, "
        repr_str += f"bbox_clip_border={self.bbox_clip_border})"
        return repr_str

    @staticmethod
    def _get_rotation_matrix(rotate_degrees: float) -> np.ndarray:
        radian = math.radians(rotate_degrees)
        rotation_matrix = np.array(
            [[np.cos(radian), -np.sin(radian), 0.0], [np.sin(radian), np.cos(radian), 0.0], [0.0, 0.0, 1.0]],
            dtype=np.float32,
        )
        return rotation_matrix

    @staticmethod
    def _get_scaling_matrix(scale_ratio: float) -> np.ndarray:
        scaling_matrix = np.array(
            [[scale_ratio, 0.0, 0.0], [0.0, scale_ratio, 0.0], [0.0, 0.0, 1.0]], dtype=np.float32
        )
        return scaling_matrix

    @staticmethod
    def _get_shear_matrix(x_shear_degrees: float, y_shear_degrees: float) -> np.ndarray:
        x_radian = math.radians(x_shear_degrees)
        y_radian = math.radians(y_shear_degrees)
        shear_matrix = np.array(
            [[1, np.tan(x_radian), 0.0], [np.tan(y_radian), 1, 0.0], [0.0, 0.0, 1.0]], dtype=np.float32
        )
        return shear_matrix

    @staticmethod
    def _get_translation_matrix(x: float, y: float) -> np.ndarray:
        translation_matrix = np.array([[1, 0.0, x], [0.0, 1, y], [0.0, 0.0, 1.0]], dtype=np.float32)
        return translation_matrix



_base_ = ["../schedule_1x.py", "../default_runtime.py", "./yolox_tta.py"]

img_scale = (640, 640)  # width, height

# model settings
model = dict(
    type="YOLOX",
    data_preprocessor=dict(
        type="DetDataPreprocessor",
        pad_size_divisor=32,
        batch_augments=[
            dict(type="BatchSyncRandomResize", random_size_range=(480, 800), size_divisor=32, interval=10)
        ],
    ),
    backbone=dict(
        type="CSPDarknet",
        deepen_factor=0.33,
        widen_factor=0.5,
        out_indices=(2, 3, 4),
        use_depthwise=False,
        spp_kernal_sizes=(5, 9, 13),
        norm_cfg=dict(type="BN", momentum=0.03, eps=0.001),
        act_cfg=dict(type="Swish"),
    ),
    neck=dict(
        type="YOLOXPAFPN",
        in_channels=[128, 256, 512],
        out_channels=128,
        num_csp_blocks=1,
        use_depthwise=False,
        upsample_cfg=dict(scale_factor=2, mode="nearest"),
        norm_cfg=dict(type="BN", momentum=0.03, eps=0.001),
        act_cfg=dict(type="Swish"),
    ),
    bbox_head=dict(
        type="YOLOXHead",
        num_classes=80,
        in_channels=128,
        feat_channels=128,
        stacked_convs=2,
        strides=(8, 16, 32),
        use_depthwise=False,
        norm_cfg=dict(type="BN", momentum=0.03, eps=0.001),
        act_cfg=dict(type="Swish"),
        loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=True, reduction="sum", loss_weight=1.0),
        loss_bbox=dict(type="IoULoss", mode="square", eps=1e-16, reduction="sum", loss_weight=5.0),
        loss_obj=dict(type="CrossEntropyLoss", use_sigmoid=True, reduction="sum", loss_weight=1.0),
        loss_l1=dict(type="L1Loss", reduction="sum", loss_weight=1.0),
    ),
    train_cfg=dict(assigner=dict(type="SimOTAAssigner", center_radius=2.5)),
    # In order to align the source code, the threshold of the val phase is
    # 0.01, and the threshold of the test phase is 0.001.
    test_cfg=dict(score_thr=0.01, nms=dict(type="nms", iou_threshold=0.65)),
)

loading_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="LoadAnnotations", with_bbox=True),
]

multi_image_mix_dataset = dict(
    mosaic=dict(
        img_scale=img_scale,
        center_ratio_range=(0.5, 1.5),
        bbox_clip_border=True,
        pad_val=114.0,
        prob=0.5,
    ),
    # TODO: add random affine    dict(
    # RandomAffine=dict(
    #     scaling_ratio_range=(0.1, 2),
    #     # img_scale is (width, height)
    #     border=(-img_scale[0] // 2, -img_scale[1] // 2),
    # ),
    mixup=dict(
        img_scale=img_scale,
        ratio_range=(0.8, 1.6),
        flip_ratio=0.5,
        pad_val=114.0,
        max_iters=15,
        bbox_clip_border=True,
    ),
)

train_pipeline = [
    dict(type="YOLOXHSVRandomAug"),
    dict(type="RandomFlip", prob=0.5),
    # According to the official implementation, multi-scale
    # training is not considered here but in the
    # 'mmdet/models/detectors/yolox.py'.
    # Resize and Pad are for the last 15 epochs when Mosaic,
    # RandomAffine, and MixUp are closed by YOLOXModeSwitchHook.
    dict(type="Resize", scale=img_scale, keep_ratio=True),
    dict(
        type="Pad",
        pad_to_square=True,
        # If the image is three-channel, the pad value needs
        # to be set separately for each channel.
        pad_val=dict(img=(114.0, 114.0, 114.0)),
    ),
    dict(type="FilterAnnotations", min_gt_bbox_wh=(1, 1), keep_empty=False),
    dict(type="PackDetInputs"),
]

test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="Resize", scale=img_scale, keep_ratio=True),
    dict(type="Pad", pad_to_square=True, pad_val=dict(img=(114.0, 114.0, 114.0))),
    dict(type="LoadAnnotations", with_bbox=True),
    dict(type="PackDetInputs", meta_keys=("img_id", "img_path", "ori_shape", "img_shape", "scale_factor")),
]



tta_model = dict(type="DetTTAModel", tta_cfg=dict(nms=dict(type="nms", iou_threshold=0.65), max_per_img=100))

img_scales = [(640, 640), (320, 320), (960, 960)]
tta_pipeline = [
    dict(type="LoadImageFromFile", backend_args=None),
    dict(
        type="TestTimeAug",
        transforms=[
            [dict(type="Resize", scale=s, keep_ratio=True) for s in img_scales],
            [
                # ``RandomFlip`` must be placed before ``Pad``, otherwise
                # bounding box coordinates after flipping cannot be
                # recovered correctly.
                dict(type="RandomFlip", prob=1.0),
                dict(type="RandomFlip", prob=0.0),
            ],
            [
                dict(type="Pad", pad_to_square=True, pad_val=dict(img=(114.0, 114.0, 114.0))),
            ],
            [dict(type="LoadAnnotations", with_bbox=True)],
            [
                dict(
                    type="PackDetInputs",
                    meta_keys=(
                        "img_id",
                        "img_path",
                        "ori_shape",
                        "img_shape",
                        "scale_factor",
                        "flip",
                        "flip_direction",
                    ),
                )
            ],
        ],
    ),
]



import logging
from typing import Callable, Optional, Union

import lightning.pytorch as pl
import torchmetrics
from lightning.pytorch.utilities import grad_norm
from torchmetrics.aggregation import BaseAggregator

from ..constants import BBOX, IMAGE, LABEL
from .lr import apply_layerwise_lr_decay, apply_single_lr, apply_two_stages_lr, get_lr_scheduler
from .utils import get_optimizer, remove_parameters_without_grad

try:
    import mmdet
    from mmcv import ConfigDict
except ImportError as e:
    mmdet = None
    ConfigDict = None

logger = logging.getLogger(__name__)


class MMDetLitModule(pl.LightningModule):
    def __init__(
        self,
        model,
        optim_type: Optional[str] = None,
        lr_choice: Optional[str] = None,
        lr_schedule: Optional[str] = None,
        lr: Optional[float] = None,
        lr_decay: Optional[float] = None,
        end_lr: Optional[Union[float, int]] = None,
        lr_mult: Optional[Union[float, int]] = None,
        weight_decay: Optional[float] = None,
        warmup_steps: Optional[int] = None,
        validation_metric: Optional[torchmetrics.Metric] = None,
        validation_metric_name: Optional[str] = None,
        custom_metric_func: Callable = None,
        test_metric: Optional[torchmetrics.Metric] = None,
        track_grad_norm: Optional[Union[int, str]] = -1,
    ):
        super().__init__()  # TODO: inherit LitModule
        self.save_hyperparameters(
            ignore=[
                "model",
                "validation_metric",
                "test_metric",
            ]
        )
        self.model = model
        self.validation_metric = validation_metric
        self.validation_metric_name = f"val_{validation_metric_name}"
        self.use_loss = isinstance(validation_metric, BaseAggregator)
        self.id2label = self.model.id2label
        self.input_data_key = self.model.prefix + "_" + IMAGE
        self.input_label_key = self.model.prefix + "_" + LABEL
        self.track_grad_norm = track_grad_norm

    def _base_step(self, batch, mode):
        ret = self.model(batch=batch[self.input_data_key], mode=mode)

        return ret

    def _predict_step(self, batch):
        return self._base_step(batch=batch, mode="predict")

    def _loss_step(self, batch):
        return self._base_step(batch=batch, mode="loss")

    def _get_map_input(self, pred_results):
        preds = []
        target = []

        batch_size = len(pred_results)

        for i in range(batch_size):
            if hasattr(pred_results[i][BBOX], "masks"):
                # has one additional dimension with 2 outputs: img_result=img_result[0], mask_result=img_result[1]
                raise NotImplementedError(
                    "Do not support training for models with masks like mask r-cnn, "
                    "because most custom datasets do not have a ground truth mask."
                    " However, you can still inference with this model."
                )
            preds.append(
                dict(
                    boxes=pred_results[i][BBOX].bboxes,  # .float().to(self.device)?
                    scores=pred_results[i][BBOX].scores,  # .float().to(self.device)?
                    labels=pred_results[i][BBOX].labels,  # .long().to(self.device)?
                )
            )
            target.append(
                dict(
                    boxes=pred_results[i][LABEL].bboxes,
                    labels=pred_results[i][LABEL].labels,
                )
            )

        return preds, target

    def evaluate(self, sample, stage=None):
        """
        sample: dict
            Single data sample.
        """
        pred_results = self._predict_step(sample)

        preds, target = self._get_map_input(pred_results)

        # use MeanAveragePrecision, example code: https://github.com/Lightning-AI/metrics/blob/master/examples/detection_map.py
        self.validation_metric.update(preds, target)

        return pred_results

    def sum_and_log_step_results(self, losses, logging=True):
        # losses is a dict of several type of losses, e.g. ['loss_cls', 'loss_conf', 'loss_xy', 'loss_wh']
        # each type of losses may have multiple channels due to multiple resolution settings
        total_loss = 0.0
        for loss_key, loss_values in losses.items():
            curr_loss = 0.0
            if isinstance(loss_values, list) or isinstance(loss_values, tuple):  # is a collection of shape 0 tensors
                for loss_chanel_idx, loss_val in enumerate(loss_values):
                    if logging:
                        self.log(f"step/{loss_key}_{loss_chanel_idx}", loss_val)
                    curr_loss += loss_val
            else:  # is a tensor
                curr_loss += loss_values.sum()

            if logging:
                self.log(f"step/{loss_key}", curr_loss)
            total_loss += curr_loss

        return total_loss

    def training_step(self, batch, batch_idx):
        losses = self._loss_step(batch=batch)
        # sum and log step losses
        total_loss = self.sum_and_log_step_results(losses)
        return total_loss

    def validation_step(self, batch, batch_idx, dataloader_idx=0):
        if self.use_loss:
            losses = self._loss_step(batch=batch)
            total_loss = self.sum_and_log_step_results(losses, logging=False)
            self.validation_metric.update(total_loss)
        else:
            self.evaluate(batch, "val")

    def on_validation_epoch_end(self):
        val_result = self.validation_metric.compute()
        if self.use_loss:
            self.log_dict({"val_direct_loss": val_result}, sync_dist=True)
        else:
            # TODO: add mAP/mAR_per_class
            val_result.pop("classes", None)  # introduced in torchmetrics v1.0.0
            mAPs = {"val_" + k: v for k, v in val_result.items()}
            mAPs["val_mAP"] = mAPs["val_map"]
            self.log_dict(mAPs, sync_dist=True)
        self.validation_metric.reset()

    def test_step(self, batch, batch_idx, dataloader_idx=0):
        raise NotImplementedError("test with lit_mmdet is not implemented yet.")

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        pred = self._predict_step(batch)

        return pred

    def configure_optimizers(self):
        """
        Configure optimizer. This function is registered by LightningModule.
        Refer to https://lightning.ai/docs/pytorch/stable/common/lightning_module.html#configure-optimizers
        Returns
        -------
        [optimizer]
            Optimizer.
        [sched]
            Learning rate scheduler.
        """
        kwargs = dict(
            model=self.model,
            lr=self.hparams.lr,
            weight_decay=self.hparams.weight_decay,
        )
        if self.hparams.lr_choice == "two_stages":
            logger.debug("applying 2-stage learning rate...")
            grouped_parameters = apply_two_stages_lr(
                lr_mult=self.hparams.lr_mult,
                return_params=True,
                **kwargs,
            )
        elif self.hparams.lr_choice == "layerwise_decay":
            logger.debug("applying layerwise learning rate decay...")
            grouped_parameters = apply_layerwise_lr_decay(
                lr_decay=self.hparams.lr_decay,
                **kwargs,
            )
        else:
            logger.debug("applying single learning rate...")
            grouped_parameters = apply_single_lr(
                **kwargs,
            )

        grouped_parameters = remove_parameters_without_grad(grouped_parameters=grouped_parameters)

        optimizer = get_optimizer(
            optim_type=self.hparams.optim_type,
            optimizer_grouped_parameters=grouped_parameters,
            lr=self.hparams.lr,
            weight_decay=self.hparams.weight_decay,
        )

        logger.debug(f"trainer.max_steps: {self.trainer.max_steps}")
        if self.trainer.max_steps is None or -1:
            max_steps = (
                len(self.trainer.datamodule.train_dataloader())
                * self.trainer.max_epochs
                // self.trainer.accumulate_grad_batches
            )
            logger.debug(
                f"len(trainer.datamodule.train_dataloader()): " f"{len(self.trainer.datamodule.train_dataloader())}"
            )
            logger.debug(f"trainer.max_epochs: {self.trainer.max_epochs}")
            logger.debug(f"trainer.accumulate_grad_batches: {self.trainer.accumulate_grad_batches}")
        else:
            max_steps = self.trainer.max_steps

        logger.debug(f"max steps: {max_steps}")

        warmup_steps = self.hparams.warmup_steps
        if isinstance(warmup_steps, float):
            warmup_steps = int(max_steps * warmup_steps)

        logger.debug(f"warmup steps: {warmup_steps}")
        logger.debug(f"lr_schedule: {self.hparams.lr_schedule}")

        scheduler = get_lr_scheduler(
            optimizer=optimizer,
            num_max_steps=max_steps,
            num_warmup_steps=warmup_steps,
            lr_schedule=self.hparams.lr_schedule,
            end_lr=self.hparams.end_lr,
        )

        # TODO: add lr_interval into hyperparameters?
        if self.hparams.lr_schedule == "multi_step":
            lr_interval = "epoch"
        else:
            lr_interval = "step"

        sched = {"scheduler": scheduler, "interval": lr_interval}
        logger.debug("done configuring optimizer and scheduler")
        return [optimizer], [sched]

    def on_before_optimizer_step(self, optimizer):
        # If using mixed precision, the gradients are already unscaled here
        if self.track_grad_norm != -1:
            self.log_dict(grad_norm(self, norm_type=self.track_grad_norm))



import torch
import torch.nn as nn


class BBCEWithLogitLoss(nn.Module):
    """
    Balanced BCEWithLogitLoss based on https://github.com/NiFangBaAGe/Explicit-Visual-Prompt/blob/latest_branch/models/segformer.py
    """

    def __init__(self):
        super(BBCEWithLogitLoss, self).__init__()

    def forward(self, input: torch.Tensor, target: torch.Tensor):
        if input.dim() == 3:
            input = input.unsqueeze(1)
        eps = 1e-10
        count_pos = torch.sum(target) + eps
        count_neg = torch.sum(1.0 - target)
        ratio = count_neg / count_pos
        w_neg = count_pos / (count_pos + count_neg)

        bce1 = nn.BCEWithLogitsLoss(pos_weight=ratio)
        loss = w_neg * bce1(input, target)

        return loss



import torch
from torchmetrics import Metric
from torchmetrics.utilities import dim_zero_cat


class Coverage(Metric):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_state("pos_probs", default=[], dist_reduce_fx="cat")
        self.add_state("targets", default=[], dist_reduce_fx="cat")
        self.tp_threshold = 0.97
        self.tn_threshold = 0.99
        higher_is_better = True

    def update(self, preds: torch.Tensor, target: torch.Tensor) -> None:
        assert preds.dim() == 1
        assert target.dim() == 1
        self.pos_probs.append(preds)
        self.targets.append(target)

    def compute(self):
        # parse inputs
        pos_probs = dim_zero_cat(self.pos_probs)
        targets = dim_zero_cat(self.targets)
        y_pos = targets[torch.where(pos_probs >= self.tp_threshold)]
        y_neg = targets[torch.where(pos_probs <= 1 - self.tn_threshold)]
        tp = sum(y_pos == 1)
        tn = sum(y_neg == 0)
        if len(y_pos) == 0 or len(targets) == 0:
            tp_precision, tp_coverage = 0, 0
        else:
            tp_precision, tp_coverage = tp / len(y_pos), len(y_pos) / len(targets)
            if tp_precision < self.tp_threshold:
                tp_coverage = 0
        if len(y_neg) == 0 or len(targets) == 0:
            tn_precision, tn_coverage = 0, 0
        else:
            tn_precision, tn_coverage = tn / len(y_neg), len(y_neg) / len(targets)
            if tn_precision < self.tn_threshold:
                tn_coverage = 0

        return torch.tensor(tp_coverage) + torch.tensor(tn_coverage)



import logging
from typing import Dict, List, Optional, Union

import numpy as np
import PIL
from PIL import ImageFile
from torch import nn

from ..infer_types import is_rois_input
from .process_mmlab_base import MMLabProcessor

try:
    from mmcv.transforms import Compose
except ImportError as e:
    pass

logger = logging.getLogger(__name__)
ImageFile.LOAD_TRUNCATED_IMAGES = True


class MMDetProcessor(MMLabProcessor):
    """
    Prepare rois data for mmdetection models.
    """

    def __init__(
        self,
        model: nn.Module,
        max_img_num_per_col: Optional[int] = 1,
        missing_value_strategy: Optional[str] = "zero",
        requires_column_info: bool = False,
    ):
        """
        Parameters
        ----------
        model
            The model using this data processor.
        max_img_num_per_col
            The maximum number of images one sample can have.
        missing_value_strategy
            How to deal with a missing image. We now support:
            - skip
                Skip this sample
            -zero
                Use an image with zero pixels.
        requires_column_info
            Whether to require feature column information in dataloader.
        """
        from ...utils import CollateMMDet, check_if_packages_installed

        check_if_packages_installed(package_names=["mmcv", "mmengine", "mmdet"])

        super().__init__(
            model=model,
            collate_func=CollateMMDet,
            max_img_num_per_col=max_img_num_per_col,
            missing_value_strategy=missing_value_strategy,
            requires_column_info=requires_column_info,
        )

        # for yolox we separate loading pipeline to support multi_image_mix_dataset
        if "loading_pipeline" in self.cfg.keys():
            self.load_processor = Compose(self.cfg.loading_pipeline)

    def prepare_one_sample(
        self,
        image_paths: Dict[str, List[str]],
        is_training: bool,
    ):
        mm_data = dict(img_prefix=None, bbox_fields=[])

        for per_col_name, per_col_content in image_paths.items():
            if is_rois_input(per_col_content):
                rois = np.array(per_col_content)
                # https://github.com/open-mmlab/mmdetection/blob/ecac3a77becc63f23d9f6980b2a36f86acd00a8a/mmdet/datasets/transforms/loading.py#L155
                mm_data["instances"] = []
                for roi in rois:
                    mm_data["instances"].append(
                        {
                            "bbox": roi[:4],
                            "bbox_label": roi[4],
                            "ignore_flag": 0,
                        }
                    )
            else:
                with PIL.Image.open(per_col_content[0]) as img:
                    # mm_data["img_info"] = dict(filename=per_col_content[0], height=img.height, width=img.width)
                    mm_data["img_path"] = per_col_content[0]
        if self.requires_column_info:
            pass  # TODO

        mm_data["img_id"] = 0  # TODO: use a non trivial image id for TTA (test time augmentation)

        return mm_data

    def load_one_sample(
        self,
        image_paths: Dict[str, List[str]],
        is_training: bool,
    ) -> Dict:
        """
        Read images, process them, and stack them. One sample can have multiple images,
        resulting in a tensor of (n, 3, size, size), where n <= max_img_num_per_col is the available image number.

        Parameters
        ----------
        image_paths
            One sample may have multiple image columns in a pd.DataFrame and multiple images
            inside each image column.
        is_training
            Whether to process images in the training mode.

        Returns
        -------
        A dictionary containing one sample's images and their number.
        """
        mm_data = self.prepare_one_sample(image_paths=image_paths, is_training=is_training)
        ret = {self.image_key: self.load_processor(mm_data)}
        return ret

    def process_one_loaded_sample(
        self,
        image_paths: Dict[str, List[str]],
        is_training: bool,
    ) -> Dict:
        """
        Read images, process them, and stack them. One sample can have multiple images,
        resulting in a tensor of (n, 3, size, size), where n <= max_img_num_per_col is the available image number.

        Parameters
        ----------
        image_paths
            One sample may have multiple image columns in a pd.DataFrame and multiple images
            inside each image column.
        is_training
            Whether to process images in the training mode.

        Returns
        -------
        A dictionary containing one sample's images and their number.
        """
        assert is_training

        image_paths.update({self.image_key: self.train_processor(image_paths[self.image_key])})

        return image_paths

    def process_one_sample(
        self,
        image_paths: Dict[str, List[str]],
        is_training: bool,
    ) -> Dict:
        """
        Read images, process them, and stack them. One sample can have multiple images,
        resulting in a tensor of (n, 3, size, size), where n <= max_img_num_per_col is the available image number.
        Parameters
        ----------
        image_paths
            One sample may have multiple image columns in a pd.DataFrame and multiple images
            inside each image column.
        is_training
            Whether to process images in the training mode.
        Returns
        -------
        A dictionary containing one sample's images and their number.
        """
        mm_data = self.prepare_one_sample(image_paths=image_paths, is_training=is_training)
        ret = {self.image_key: self.train_processor(mm_data) if is_training else self.val_processor(mm_data)}
        return ret

    def __call__(
        self,
        images: Dict[str, List[str]],
        feature_modalities: Dict[str, Union[int, float, list]],
        is_training: bool,
        load_only: bool = False,
    ) -> Dict:
        """
        Obtain one sample's images and customized them for a specific model.

        Parameters
        ----------
        images
            Images of one sample.
        feature_modalities
            The modality of the feature columns.
        is_training
            Whether to process images in the training mode.
        load_only
            Whether to only load the data. Other processing steps may happen in dataset.__getitem__.

        Returns
        -------
        A dictionary containing one sample's processed images and their number.
        """
        images = {k: [v] if isinstance(v, str) else v for k, v in images.items()}
        if load_only:
            return self.load_one_sample(images, is_training)
        else:
            return self.process_one_sample(images, is_training)



import base64
import collections
import json
import logging
import warnings
from io import BytesIO
from typing import Dict, List, Optional, Tuple, Union

import pandas as pd
import PIL
import pytesseract

from autogluon.core.utils import infer_problem_type as infer_basic_problem_type

from ..constants import (
    BINARY,
    CATEGORICAL,
    CLASSIFICATION,
    DOCUMENT_IMAGE,
    DOCUMENT_PDF,
    IDENTIFIER,
    IMAGE_BASE64_STR,
    IMAGE_BYTEARRAY,
    IMAGE_PATH,
    MULTICLASS,
    NER,
    NER_ANNOTATION,
    NULL,
    NUMERICAL,
    OBJECT_DETECTION,
    REGRESSION,
    ROIS,
    SEMANTIC_SEGMENTATION,
    SEMANTIC_SEGMENTATION_GT,
    SEMANTIC_SEGMENTATION_IMG,
    TEXT,
    TEXT_NER,
)

logger = logging.getLogger(__name__)


def is_categorical_column(
    data: pd.Series,
    valid_data: pd.Series,
    threshold: int = None,
    ratio: Optional[float] = None,
    oov_ratio_threshold: Optional[float] = None,
    is_label: bool = False,
) -> bool:
    """
    Identify whether a column is one categorical column.
    If the number of unique elements in the column is smaller than

        min(#Total Sample * ratio, threshold),

    it will be treated as a categorical column.

    Parameters
    ----------
    data
        One column of a multimodal pd.DataFrame for training.
    valid_data
        One column of a multimodal pd.DataFrame for validation.
    threshold
        The threshold for detecting categorical column.
    ratio
        The ratio detecting categorical column.
    oov_ratio_threshold
        The out-of-vocabulary ratio between training and validation.
        This is used to determine if the column is a categorical column.
        Usually, a categorical column can tolerate a small OOV ratio.
    is_label
        Whether the column is a label column.

    Returns
    -------
    Whether the column is a categorical column.
    """
    if data.dtype.name == "category":
        return True
    else:
        if threshold is None:
            if is_label:
                # TODO(?) The following logic will be problematic if the task is few-shot learning.
                threshold = 100
                oov_ratio_threshold = 0
                ratio = 0.1
            else:
                threshold = 20
                oov_ratio_threshold = 0
                ratio = 0.1
        threshold = min(int(len(data) * ratio), threshold)
        try:
            data_value_counts = data.value_counts(dropna=False)
            key_set = set(data_value_counts.keys())
            if len(data_value_counts) < threshold:
                valid_value_counts = valid_data.value_counts(dropna=False)
                total_valid_num = len(valid_data)
                oov_num = 0
                for k, v in zip(valid_value_counts.keys(), valid_value_counts.values):
                    if k not in key_set:
                        oov_num += v
                if is_label and oov_num != 0:
                    return False
                if oov_num / total_valid_num > oov_ratio_threshold:
                    return False
                return True
            else:
                return False
        except:
            return False


def is_rois_input(sample):
    """
    check if a sample is rois for object detection

    Parameters
    ----------
    sample
        The sampled data.

    Returns
    -------
    bool, whether a sample is rois for object detection
    """
    return isinstance(sample, list) and len(sample) and isinstance(sample[0], list) and len(sample[0]) == 5


def is_rois_column(data: pd.Series) -> bool:
    """
    Identify if a column is one rois column.

    Parameters
    ----------
    X
        One column of a multimodal pd.DataFrame for training.

    Returns
    -------
    Whether the column is a rois column.
    """
    idx = data.first_valid_index()
    if isinstance(data[idx], str):
        try:
            rois = json.loads(data[idx][0])
            # TODO: better infer logic / input format to not confuse with other modality
            return rois and isinstance(rois, list) and len(data[idx][0]) == 5
        except:
            return False
    else:
        return is_rois_input(data[idx])


def is_numerical_column(
    data: pd.Series,
    valid_data: Optional[pd.Series] = None,
) -> bool:
    """
    Identify if a column is a numerical column.
    Here it uses a very simple rule to verify if this is a numerical column.

    Parameters
    ----------
    data
        One column of a multimodal pd.DataFrame for training.
    valid_data
        One column of a multimodal pd.DataFrame for validation.

    Returns
    -------
    Whether the column is a numerical column.
    """
    try:
        numerical_data = pd.to_numeric(data)
        if valid_data is not None:
            numerical_valid_data = pd.to_numeric(valid_data)
        return True
    except:
        return False


def is_image_column(
    data: pd.Series,
    col_name: str,
    image_type: str,
    sample_n: Optional[int] = 500,
) -> bool:
    """
    Identify if a column is one image-path column.
    Here it counts the failures when trying PIL.Image.open() on a sampled subset.
    If over 90% attempts fail, this column isn't an image-path column.

    Parameters
    ----------
    data
        One column of a multimodal pd.DataFrame for training.
    col_name
        Name of column.
    image_type
        The image type to check.
    sample_n
        Number of sample images to open for sanity check.

    Returns
    -------
    Whether the column is an image-path column.
    """
    sample_num = min(len(data), sample_n)
    data = data.sample(n=sample_num, random_state=0)
    if image_type == IMAGE_PATH:
        data = data.apply(lambda ele: str(ele).split(";")).tolist()
    elif image_type in [IMAGE_BYTEARRAY, IMAGE_BASE64_STR]:
        data = data.tolist()
    else:
        raise ValueError(f"Unsupported image type: {image_type}")

    failure_count = 0
    for images in data:
        success = False
        if not isinstance(images, list):
            images = [images]
        for per_image in images:
            try:
                if image_type == IMAGE_PATH:
                    with PIL.Image.open(per_image) as img:
                        pass
                elif image_type == IMAGE_BYTEARRAY:
                    with PIL.Image.open(BytesIO(per_image)) as img:
                        pass
                elif image_type == IMAGE_BASE64_STR:
                    with PIL.Image.open(BytesIO(base64.b64decode(per_image))) as img:
                        pass
                else:
                    raise ValueError(f"Unsupported image type: {image_type}")
            except:
                success = False
                break

            success = True
        if not success:
            failure_count += 1
    failure_ratio = failure_count / sample_num
    # Tolerate high failure rate in case that many image files may be corrupted.
    if failure_ratio <= 0.9:
        if failure_ratio > 0:
            warnings.warn(
                f"Among {sample_num} sampled images in column '{col_name}', "
                f"{failure_ratio:.0%} images can't be open. "
                "You may need to thoroughly check your data to see the percentage of missing images, "
                "and estimate the potential influence. By default, we use an image with zero pixels. "
                "You can also set hyperparameter 'data.image.missing_value_strategy' to be 'skip', "
                "which skips samples that contain a missing image.",
                UserWarning,
            )
        return True
    else:
        return False


def is_document_pdf_column(
    data: pd.Series,
    col_name: str,
    sample_m: Optional[int] = 500,
) -> bool:
    """
    Identify if a column is a pdf document column.

    Parameters
    ----------
    data
        One column of a multimodal pd.DataFrame for training.
    col_name
        Name of column.
    sample_m
        Number of sample images used to check if images are documents images.

    Returns
    -------
    Whether the column is a pdf document column.
    """

    if len(data) > sample_m:
        # Sample to speed-up type inference
        data = data.sample(n=sample_m, random_state=0)

    for docs in data:
        try:
            if not isinstance(docs, list):
                docs = [docs]
            for per_doc in docs:
                # If there is non-pdf document, return False
                if not per_doc.endswith(".pdf"):
                    return False
        except:
            return False

    return True


def is_document_image_column(
    data: pd.Series,
    col_name: str,
    image_type: Optional[str] = IMAGE_PATH,
    sample_m: Optional[int] = 10,
    text_len_threshold: Optional[int] = 100,
) -> bool:
    """
    Identify if a column is a document image column.

    Parameters
    ----------
    data
        One column of a multimodal pd.DataFrame for training.
    col_name
        Name of column.
    image_type
        The image type to check. Set to IMAGE_PATH by default.
    sample_m
        Number of sample images used to check if images are documents images.
    text_len_threshold
        If the average text length is longer than text_len_threshold, the images will be considered as document images.
    Returns
    -------
    Whether the column is a document image column.
    """

    # TODO: Add support for other types (e.g., pdf) of document.

    words_len = []
    if len(data) > sample_m:
        # Sample to speed-up type inference
        data = data.sample(n=sample_m, random_state=0)
    failure_count = 0
    for images in data:
        success = False
        if not isinstance(images, list):
            images = [images]
        for per_image in images:
            try:
                # convert images to string
                with PIL.Image.open(per_image) as doc_image:
                    words = pytesseract.image_to_string(doc_image)
                    words_len.append(len(words))
            except Exception as e:
                words_len.append(0)
                success = False
                break
            success = True
        if not success:
            failure_count += 1

    if (1 - failure_count / sample_m) >= 0.8:
        logger.debug(f"Average length of words of this dataset is {sum(words_len) / len(words_len)}.")
        if sum(words_len) / len(words_len) > text_len_threshold:
            return True
        else:
            return False
    else:
        False


def is_text_column(data: pd.Series) -> bool:
    """
    Identify if a column is one text column.

    Parameters
    ----------
    data
        One column of a multimodal pd.DataFrame for training.

    Returns
    -------
    Whether the column is a text column.
    """
    if len(data) > 5000:
        # Sample to speed-up type inference
        data = data.sample(n=5000, random_state=0)
    try:
        data_unique = data.unique()
        num_unique = len(data_unique)
        num_rows = len(data)
        unique_ratio = num_unique / num_rows
        if unique_ratio <= 0.01:
            return False
        try:
            avg_words = pd.Series(data_unique).str.split().str.len().mean()
        except AttributeError:
            return False
        if avg_words < 3:
            return False
    except:
        return False
    return True


def is_identifier_column(data: pd.Series, col_name: str, id_mappings: Dict[str, Dict]) -> bool:
    """
    Check if a column is one identifier column.

    Parameters
    ----------
    data
        One column of multimodal pd.DataFrame.
    col_name
        Name of the column.
    id_mappings
        Id-to-content mappings. The contents can be text, image, etc.
        This is used when the dataframe contains the query/response indexes instead of their contents.

    Returns
    -------
    Whether the column is an identifier column.
    """
    if not id_mappings or col_name not in id_mappings:
        return False

    sample_num = min(len(data), 500)
    data = data.sample(n=sample_num, random_state=0).tolist()
    failure_count = 0
    for index in data:
        try:
            per_value = id_mappings[col_name][index]
        except:
            failure_count += 1

    if failure_count == 0:
        return True
    elif 0 < failure_count < sample_num:
        warnings.warn(
            f"Among {sample_num} sampled indexes in column {col_name}, "
            f"we can't index all their values from the id_mappings ({failure_count} failures). "
            f"You may need to assure that all the indexes of column {col_name} exist in your id_mappings.",
            UserWarning,
        )
    else:
        return False


def infer_id_mappings_types(id_mappings: Union[Dict[str, Dict], Dict[str, pd.Series]]) -> Dict:
    """
    Infer the data types in id_mappings.

    Parameters
    ----------
    id_mappings
        Id-to-content mappings. The contents can be text, image, etc.

    Returns
    -------
    A dictionary containing the data types in id_mappings.
    """
    id_mappings_types = collections.OrderedDict()
    if id_mappings is None:
        return id_mappings_types

    for per_name, per_id_mappings in id_mappings.items():
        if isinstance(per_id_mappings, dict):
            per_id_mappings = pd.Series(per_id_mappings.values())
        elif isinstance(per_id_mappings, pd.Series):
            pass
        else:
            raise ValueError(
                f"Invalid per_id_mappings type: {type(per_id_mappings)}. Make sure the id_mappings is a dict of dicts or a dict of pd.Series."
            )
        if is_image_column(per_id_mappings, col_name=per_name, image_type=IMAGE_PATH):
            id_mappings_types[per_name] = IMAGE_PATH
        elif is_text_column(per_id_mappings):
            id_mappings_types[per_name] = TEXT
        elif is_image_column(per_id_mappings, col_name=per_name, image_type=IMAGE_BYTEARRAY):
            id_mappings_types[per_name] = IMAGE_BYTEARRAY
        elif is_image_column(per_id_mappings, col_name=per_name, image_type=IMAGE_BASE64_STR):
            id_mappings_types[per_name] = IMAGE_BASE64_STR
        else:
            raise ValueError(
                f"{per_name} in the id_mappings has an invalid type. Currently, we only support image and text types."
            )
    return id_mappings_types


def infer_column_types(
    data: pd.DataFrame,
    valid_data: Optional[pd.DataFrame] = None,
    label_columns: Union[str, List[str]] = None,
    provided_column_types: Optional[Dict] = None,
    allowable_column_types: Optional[List[str]] = None,
    fallback_column_type: Optional[str] = None,
    id_mappings: Optional[Union[Dict[str, Dict], Dict[str, pd.Series]]] = None,
    problem_type: Optional[str] = None,
) -> Dict:
    """
    Infer the column types of a multimodal pd.DataFrame.

    Parameters
    ----------
    data
        The multimodal pd.DataFrame for training.
    valid_data
        The multimodal pd.DataFrame for validation.
    label_columns
        The label column names.
    provided_column_types
        Additional dictionary that you can use to specify the columns types that you know.
        {'col_name': TYPE, ...}
    allowable_column_types
        What column types are allowed. This is the prior knowledge inferred from the model type.
    fallback_column_type
        What's the fallback column type if the detected type if out of the allowable_column_types.
    id_mappings
        Id-to-content mappings. The contents can be text, image, etc.
        This is used when the dataframe contains the query/response indexes instead of their contents.
    problem_type
        The problem type used to update some column types.

    Returns
    -------
    column_types
        A dictionary containing the mappings from column names to their modality types.
        If the column does not contain any useful information, we will set its column type NULL.
    """
    if label_columns is None:
        label_columns = []
    if isinstance(label_columns, str):
        label_columns = [label_columns]

    column_types = collections.OrderedDict()

    if valid_data is None:
        valid_data = data
        is_training = False
    else:
        is_training = True

    id_mappings_types = infer_id_mappings_types(id_mappings)

    for col_name in data.columns:
        if provided_column_types is not None and col_name in provided_column_types:
            column_types[col_name] = provided_column_types[col_name]
            continue
        # Identify columns that provide no information
        idx = data[col_name].first_valid_index()
        if idx is None:
            # No valid index, thus, we will just ignore the column
            column_types[col_name] = NULL
            continue
        if (
            (not isinstance(data[col_name][idx], (list, dict, set, bytearray)))
            and len(data[col_name].unique()) == 1
            and is_training
        ):
            column_types[col_name] = NULL
            continue
        # TODO: valid check for collections

        if is_rois_column(data[col_name]):
            column_types[col_name] = ROIS
        # keep the elif here because ROIS need to skip the categorical check
        # where error occurs due to List type input
        elif is_identifier_column(data[col_name], col_name=col_name, id_mappings=id_mappings):
            column_types[col_name] = f"{id_mappings_types[col_name]}_{IDENTIFIER}"
        elif is_categorical_column(
            data[col_name], valid_data[col_name], is_label=col_name in label_columns
        ):  # Infer categorical column
            column_types[col_name] = CATEGORICAL
        elif is_numerical_column(data[col_name], valid_data[col_name]):  # Infer numerical column
            column_types[col_name] = NUMERICAL
        elif is_image_column(data[col_name], col_name=col_name, image_type=IMAGE_PATH):  # Infer image-path column
            # Check if it is document image or not.
            if is_document_image_column(data[col_name], col_name=col_name):
                column_types[col_name] = DOCUMENT_IMAGE
            elif problem_type == SEMANTIC_SEGMENTATION:
                column_types[col_name] = SEMANTIC_SEGMENTATION_IMG
            else:
                column_types[col_name] = IMAGE_PATH
        elif is_document_pdf_column(data[col_name], col_name=col_name):
            column_types[col_name] = DOCUMENT_PDF
        elif is_text_column(data[col_name]):  # Infer text column
            column_types[col_name] = TEXT
        elif is_image_column(
            data[col_name], col_name=col_name, image_type=IMAGE_BYTEARRAY
        ):  # Infer image-bytearray column
            column_types[col_name] = IMAGE_BYTEARRAY
        elif is_image_column(
            data[col_name], col_name=col_name, image_type=IMAGE_BASE64_STR
        ):  # Infer image-base64str column
            column_types[col_name] = IMAGE_BASE64_STR
        else:  # All the other columns are treated as categorical
            column_types[col_name] = CATEGORICAL

    if allowable_column_types and fallback_column_type:
        column_types = set_fallback_column_type(
            column_types=column_types,
            allowable_column_types=allowable_column_types,
            fallback_column_type=fallback_column_type,
        )
    if problem_type == NER:
        column_types = infer_ner_column_type(column_types=column_types)

    if problem_type and label_columns:
        column_types = infer_label_column_type_by_problem_type(
            column_types=column_types,
            label_columns=label_columns,
            problem_type=problem_type,
            data=data,
            valid_data=valid_data if is_training else None,
        )

    return column_types


def check_missing_values(
    data: pd.DataFrame,
    column_name: str,
    split: Optional[str] = "",
):
    num_missing_values = data[column_name].isnull().sum()
    if num_missing_values > 0:
        raise ValueError(
            f"Label column '{column_name}' contains missing values in the "
            f"{split} dataframe. You may want to filter your data because "
            "missing label is currently not supported."
        )


def infer_label_column_type_by_problem_type(
    column_types: Dict,
    label_columns: Union[str, List[str]],
    problem_type: Optional[str],
    data: Optional[pd.DataFrame] = None,
    valid_data: Optional[pd.DataFrame] = None,
    allowable_label_types: Optional[List[str]] = (
        CATEGORICAL,
        NUMERICAL,
        NER_ANNOTATION,
        ROIS,
        SEMANTIC_SEGMENTATION_GT,
    ),
    fallback_label_type: Optional[str] = CATEGORICAL,
):
    """
    Infer the label column types based on problem type.

    Parameters
    ----------
    column_types
        Types of columns in a pd.DataFrame.
    label_columns
        The label columns in a pd.DataFrame.
    problem_type
        Type of problem.
    data
        A pd.DataFrame.
    valid_data
        A validation pd.DataFrame.
    allowable_label_types
        Which label types are allowed.
    fallback_label_type
        If a label type is not within the allowable_label_types, replace it with this fallback_label_type.

    Returns
    -------
    Column types with the label columns' types inferred from the problem type.
    """
    if label_columns is None:
        return column_types

    if isinstance(label_columns, str):
        label_columns = [label_columns]

    for col_name in label_columns:
        # For zero-shot inference, label column is unnecessary
        if col_name not in column_types:
            continue
        if data is not None:
            check_missing_values(data=data, column_name=col_name, split="training")
        if valid_data is not None:
            check_missing_values(data=valid_data, column_name=col_name, split="validation")
        if column_types[col_name] == NULL:
            raise ValueError(
                f"Label column '{col_name}' contains only one label class. Make sure it has at least two label classes."
            )
        if problem_type in [MULTICLASS, BINARY, CLASSIFICATION]:
            column_types[col_name] = CATEGORICAL
        elif problem_type == REGRESSION:
            column_types[col_name] = NUMERICAL
        elif problem_type == NER:
            column_types[col_name] = NER_ANNOTATION
        elif problem_type == OBJECT_DETECTION:
            column_types[col_name] = ROIS
        elif problem_type == SEMANTIC_SEGMENTATION:
            column_types[col_name] = SEMANTIC_SEGMENTATION_GT

        if column_types[col_name] not in allowable_label_types:
            column_types[col_name] = fallback_label_type

    return column_types


def infer_rois_column_type(
    column_types: Dict,
    data: Optional[pd.DataFrame] = None,
):
    for col_name in column_types.keys():
        if is_rois_column(data[col_name]):
            column_types[col_name] = ROIS
    return column_types


def infer_problem_type(
    y_train_data: Optional[pd.DataFrame] = None,
    provided_problem_type: Optional[str] = None,
) -> str:
    """
    Infer the basic (binary, multiclass, and regression) problem types if problem type is None or classification.
    Only training data are used. Assume users provide valid validation data.

    Parameters
    ----------
    y_train_data
        The label column of training data.
    provided_problem_type
        Provided problem type

    Returns
    -------
    Inferred problem type
    """
    if provided_problem_type in [None, CLASSIFICATION]:
        problem_type = infer_basic_problem_type(y=y_train_data)
        # trust users' prior knowledge
        if provided_problem_type == CLASSIFICATION and problem_type == REGRESSION:
            problem_type = MULTICLASS
    else:
        problem_type = provided_problem_type

    return problem_type


def infer_output_shape(
    label_column: str,
    problem_type: str,
    data: pd.DataFrame,
) -> int:
    """
    Infer the output shape based on the label column type, training data, and problem type.
    Binary classification should have class number 2, while multi-class classification's class
    number should be larger than 2. For regression, the output is restricted to 1 scalar.

    Parameters
    ----------
    label_column
        The label column in a multimodal pd.DataFrame.
    data
        The multimodal pd.DataFrame for training.
    problem_type
        Problem type which can be inferred or provided.

    Returns
    -------
    problem_type
        Type of problem.
    output_shape
        Shape of output.
    """
    if label_column is None:
        return None

    if problem_type in [BINARY, MULTICLASS, REGRESSION, CLASSIFICATION]:
        class_num = len(data[label_column].unique())
        err_msg = (
            f"Problem type is '{problem_type}' while the number of "
            f"unique values in the label column is {class_num}."
        )
        if problem_type == BINARY:
            if class_num != 2:
                raise AssertionError(err_msg)
            return 2
        elif problem_type == MULTICLASS:
            if class_num <= 2:
                raise AssertionError(err_msg)
            return class_num
        elif problem_type == REGRESSION:
            if class_num < 1:
                raise AssertionError(err_msg)
            return 1
        else:
            return class_num
    elif problem_type in [NER, OBJECT_DETECTION, SEMANTIC_SEGMENTATION]:
        return None
    else:
        raise ValueError(
            f"Problem type '{problem_type}' doesn't have a valid output shape "
            f"for training. The supported problem types are"
            f" '{BINARY}', '{MULTICLASS}', '{REGRESSION}',"
            f" '{CLASSIFICATION}', '{NER}',"
            f" '{OBJECT_DETECTION}', "
            f" '{SEMANTIC_SEGMENTATION}"
        )


def set_fallback_column_type(column_types: Dict, allowable_column_types: List[str], fallback_column_type: str) -> Dict:
    """
    Filter the auto-detected column types to make sure that all column types are allowable.
    Use the fallback type to replace those out of the allowable_column_types.

    Parameters
    ----------
    column_types
        The inferred column types.
    allowable_column_types
        The column types which are allowed by the model type.
    fallback_column_type
        Fallback to this type if some invalid column type is found.

    Returns
    -------
    The filtered column types.
    """
    for col_name, col_type in column_types.items():
        if not col_type.startswith(tuple(allowable_column_types)):
            column_types[col_name] = fallback_column_type

    return column_types


def infer_ner_column_type(column_types: Dict):
    """
    Replace the first text with text_ner for the ner problem type
    because no text_ner is detected in infer_column_types.

    Parameters
    ----------
    column_types
        The column types of a dataframe.

    Returns
    -------
    The columns types with one text_ner inside it.
    """
    # if there is already one column has type text_ner, no action is taken.
    if any([col_type.startswith(TEXT_NER) for col_type in column_types.values()]):
        return column_types

    for column in (
        column_types.keys()
    ):  # column_types is an ordered dict, so column_types.keys() returns the keys in the order of insertions.
        if column_types[column].startswith(TEXT):
            column_types[column] = column_types[column].replace(TEXT, TEXT_NER)
            break

    return column_types



from .cache import DDPPredictionWriter
from .checkpoint import AutoMMModelCheckpoint, AutoMMModelCheckpointIO, average_checkpoints
from .config import (
    apply_omegaconf_overrides,
    customize_model_names,
    filter_hyperparameters,
    filter_search_space,
    filter_timm_pretrained_cfg,
    get_config,
    get_default_config,
    get_local_pretrained_config_paths,
    get_pretrain_configs_dir,
    parse_dotlist_conf,
    save_pretrained_model_configs,
    split_hyperparameters,
    update_config_by_rules,
    update_hyperparameters,
    update_tabular_config_by_resources,
    update_ensemble_hyperparameters,
)
from .device import compute_num_gpus, get_available_devices, move_to_device
from .distillation import DistillationMixin
from .download import download, is_url
from .export import ExportMixin
from .hpo import hyperparameter_tune
from .inference import RealtimeMixin, compute_inference_batch_size, extract_from_output
from .load import CustomUnpickler, protected_zip_extraction, get_dir_ckpt_paths, get_load_ckpt_paths
from .log import (
    LogFilter,
    apply_log_filter,
    get_gpu_message,
    on_fit_end_message,
    on_fit_per_run_start_message,
    on_fit_start_message,
)
from .matcher import compute_semantic_similarity, convert_data_for_ranking, create_siamese_model, semantic_search
from .misc import (
    logits_to_prob,
    path_expander,
    path_to_base64str_expander,
    path_to_bytearray_expander,
    shopee_dataset,
    tensor_to_ndarray,
    merge_bio_format,
)
from .mmcv import CollateMMDet, CollateMMOcr
from .object_detection import (
    COCODataset,
    bbox_ratio_xywh_to_index_xyxy,
    bbox_xyxy_to_xywh,
    cocoeval,
    convert_pred_to_xywh,
    convert_result_df,
    from_coco,
    from_coco_or_voc,
    from_dict,
    from_voc,
    get_detection_classes,
    object_detection_data_to_df,
    object_detection_df_to_coco,
    save_result_coco_format,
    save_result_voc_format,
    visualize_detection,
)
from .precision import get_precision_context, infer_precision
from .presets import get_basic_config, get_ensemble_presets, get_presets, list_presets, matcher_presets
from .problem_types import PROBLEM_TYPES_REG, infer_problem_type_by_eval_metric
from .save import process_save_path, setup_save_path, make_exp_dir
from .strategy import is_interactive_strategy, run_ddp_only_once
from .env import is_interactive_env
from .visualizer import NERVisualizer, ObjectDetectionVisualizer, SemanticSegmentationVisualizer, visualize_ner
from .install import check_if_packages_installed



from .softmax_losses import MultiNegativesSoftmaxLoss, SoftTargetCrossEntropy
from .focal_loss import FocalLoss
from .lemda_loss import LemdaLoss
from .rkd_loss import RKDLoss
from .bce_loss import BBCEWithLogitLoss
from .structure_loss import StructureLoss
from .utils import (
    generate_metric_learning_labels,
    get_aug_loss_func,
    get_loss_func,
    get_matcher_loss_func,
    get_matcher_miner_func,
    get_metric_learning_distance_func,
)



import copy
import json
import logging
import os
import pathlib
import pickle
import pprint
import time
import warnings
from typing import Callable, Dict, List, Optional, Union

import numpy as np
import pandas as pd

from autogluon.core.metrics import Scorer
from autogluon.core.models.greedy_ensemble.ensemble_selection import EnsembleSelection

from .. import version as ag_version
from ..constants import BINARY, LOGITS, MULTICLASS, REGRESSION, TEST, VAL, Y_PRED, Y_TRUE
from ..optim import compute_score
from ..utils import (
    extract_from_output,
    get_dir_ckpt_paths,
    logits_to_prob,
    on_fit_end_message,
    update_ensemble_hyperparameters,
)
from .base import BaseLearner

logger = logging.getLogger(__name__)


class EnsembleLearner(BaseLearner):
    def __init__(
        self,
        label: Optional[str] = None,
        problem_type: Optional[str] = None,
        presets: Optional[str] = "high_quality",
        eval_metric: Optional[Union[str, Scorer]] = None,
        hyperparameters: Optional[dict] = None,
        path: Optional[str] = None,
        verbosity: Optional[int] = 2,
        warn_if_exist: Optional[bool] = True,
        enable_progress_bar: Optional[bool] = None,
        ensemble_size: Optional[int] = 2,
        ensemble_mode: Optional[str] = "one_shot",
        **kwargs,
    ):
        """
        Parameters
        ----------
        label
            Name of the column that contains the target variable to predict.
        problem_type
            Type of the prediction problem. We support standard problems like

            - 'binary': Binary classification
            - 'multiclass': Multi-class classification
            - 'regression': Regression
            - 'classification': Classification problems include 'binary' and 'multiclass' classification.
        presets
            Presets regarding model quality, e.g., best_quality, high_quality, and medium_quality.
        eval_metric
            Evaluation metric name. If `eval_metric = None`, it is automatically chosen based on `problem_type`.
            Defaults to 'accuracy' for multiclass classification, `roc_auc` for binary classification, and 'root_mean_squared_error' for regression.
        hyperparameters
            This is to override some default configurations.
            For example, changing the text and image backbones can be done by formatting:

            a string
            hyperparameters = "model.hf_text.checkpoint_name=google/electra-small-discriminator model.timm_image.checkpoint_name=swin_small_patch4_window7_224"

            or a list of strings
            hyperparameters = ["model.hf_text.checkpoint_name=google/electra-small-discriminator", "model.timm_image.checkpoint_name=swin_small_patch4_window7_224"]

            or a dictionary
            hyperparameters = {
                            "model.hf_text.checkpoint_name": "google/electra-small-discriminator",
                            "model.timm_image.checkpoint_name": "swin_small_patch4_window7_224",
                        }
        path
            Path to directory where models and intermediate outputs should be saved.
            If unspecified, a time-stamped folder called "AutogluonAutoMM/ag-[TIMESTAMP]"
            will be created in the working directory to store all models.
            Note: To call `fit()` twice and save all results of each fit,
            you must specify different `path` locations or don't specify `path` at all.
            Otherwise files from first `fit()` will be overwritten by second `fit()`.
        verbosity
            Verbosity levels range from 0 to 4 and control how much information is printed.
            Higher levels correspond to more detailed print statements (you can set verbosity = 0 to suppress warnings).
            If using logging, you can alternatively control amount of information printed via `logger.setLevel(L)`,
            where `L` ranges from 0 to 50
            (Note: higher values of `L` correspond to fewer print statements, opposite of verbosity levels)
        warn_if_exist
            Whether to raise warning if the specified path already exists.
        enable_progress_bar
            Whether to show progress bar. It will be True by default and will also be
            disabled if the environment variable os.environ["AUTOMM_DISABLE_PROGRESS_BAR"] is set.
        ensemble_size
            A multiple of number of models in the ensembling pool (Default 2). The actual ensemble size = ensemble_size * the model number
        ensemble_mode
            The mode of conducting ensembling:
            - `one_shot`: the classic ensemble selection
            - `sequential`: iteratively calling the classic ensemble selection with each time growing the model zoo by the best next model.
        """
        super().__init__(
            label=label,
            problem_type=problem_type,
            presets=presets,
            eval_metric=eval_metric,
            hyperparameters=hyperparameters,
            path=path,
            verbosity=verbosity,
            warn_if_exist=warn_if_exist,
            enable_progress_bar=enable_progress_bar,
        )
        self._ensemble_size = int(ensemble_size)
        assert ensemble_mode in ["sequential", "one_shot"]
        self._ensemble_mode = ensemble_mode
        self._weighted_ensemble = None
        self._selected_learners = None
        self._all_learners = None
        self._selected_indices = []
        self._relative_path = False

        return

    def get_learner_path(self, learner_path: str):
        if self._relative_path:
            learner_path = os.path.join(self._save_path, learner_path)
        return learner_path

    def get_learner_name(self, learner):
        if isinstance(learner, str):
            if self._relative_path:
                learner_name = learner
            else:
                learner_name = pathlib.PurePath(learner).name
        else:
            learner_name = pathlib.PurePath(learner.path).name

        return learner_name

    def predict_all_for_ensembling(
        self,
        learners: List[Union[str, BaseLearner]],
        data: Union[pd.DataFrame, str],
        mode: str,
        requires_label: Optional[bool] = False,
        save: Optional[bool] = False,
    ):
        assert mode in [VAL, TEST]
        predictions = []
        labels = None
        i = 0
        for per_learner in learners:
            i += 1
            logger.info(f"\npredicting with learner {i}: {per_learner}\n")
            if isinstance(per_learner, str):
                per_learner_path = self.get_learner_path(per_learner)
            else:
                per_learner_path = per_learner.path

            pred_file_path = os.path.join(per_learner_path, f"{mode}_predictions.npy")
            if os.path.isfile(pred_file_path):
                logger.info(f"{mode}_predictions.npy exists. loading it...")
                y_pred = np.load(pred_file_path)
            else:
                if isinstance(per_learner, str):
                    per_learner = BaseLearner.load(path=per_learner_path)
                if not self._problem_type:
                    self._problem_type = per_learner.problem_type
                else:
                    assert self._problem_type == per_learner.problem_type
                outputs = per_learner.predict_per_run(
                    data=data,
                    realtime=False,
                    requires_label=False,
                )
                y_pred = extract_from_output(outputs=outputs, ret_type=LOGITS)

                if self._problem_type == REGRESSION:
                    y_pred = per_learner._df_preprocessor.transform_prediction(y_pred=y_pred)
                if self._problem_type in [BINARY, MULTICLASS]:
                    y_pred = logits_to_prob(y_pred)
                    if self._problem_type == BINARY:
                        y_pred = y_pred[:, 1]

                if save:
                    np.save(pred_file_path, y_pred)

            if requires_label:
                label_file_path = os.path.join(per_learner_path, f"{mode}_labels.npy")
                if os.path.isfile(label_file_path):
                    logger.info(f"{mode}_labels.npy exists. loading it...")
                    y_true = np.load(label_file_path)
                else:
                    if isinstance(per_learner, str):
                        per_learner = BaseLearner.load(path=per_learner_path)
                    y_true = per_learner._df_preprocessor.transform_label_for_metric(df=data)

                    if save:
                        np.save(label_file_path, y_true)

                if labels is None:
                    labels = y_true
                else:
                    assert np.array_equal(y_true, labels)

            predictions.append(y_pred)

        if requires_label:
            return predictions, labels
        else:
            return predictions

    @staticmethod
    def verify_predictions_labels(predictions, learners, labels=None):
        if labels is not None:
            assert isinstance(labels, np.ndarray)
        assert isinstance(predictions, list) and all(isinstance(ele, np.ndarray) for ele in predictions)
        assert len(learners) == len(
            predictions
        ), f"len(learners) {len(learners)} doesn't match len(predictions) {len(predictions)}"

    def fit_per_ensemble(
        self,
        predictions: List[np.ndarray],
        labels: np.ndarray,
    ):
        weighted_ensemble = EnsembleSelection(
            ensemble_size=self._ensemble_size * len(predictions),
            problem_type=self._problem_type,
            metric=self._eval_metric_func,
        )
        weighted_ensemble.fit(predictions=predictions, labels=labels)

        return weighted_ensemble

    def select_next_best(self, left_learner_indices, selected_learner_indices, predictions, labels):
        best_regret = None
        best_weighted_ensemble = None
        best_next_index = None
        for i in left_learner_indices:
            tmp_learner_indices = selected_learner_indices + [i]
            tmp_predictions = [predictions[j] for j in tmp_learner_indices]
            tmp_weighted_ensemble = self.fit_per_ensemble(
                predictions=tmp_predictions,
                labels=labels,
            )
            if best_regret is None or tmp_weighted_ensemble.train_score_ < best_regret:
                best_regret = tmp_weighted_ensemble.train_score_
                best_weighted_ensemble = tmp_weighted_ensemble
                best_next_index = i

        return best_regret, best_next_index, best_weighted_ensemble

    def sequential_ensemble(
        self,
        predictions: List[np.ndarray],
        labels: np.ndarray,
    ):
        selected_learner_indices = []
        all_learner_indices = list(range(len(predictions)))
        best_regret = None
        best_weighted_ensemble = None
        best_selected_learner_indices = None
        while len(selected_learner_indices) < len(all_learner_indices):
            left_learner_indices = [i for i in all_learner_indices if i not in selected_learner_indices]
            assert sorted(all_learner_indices) == sorted(selected_learner_indices + left_learner_indices)
            logger.debug(f"\nleft_learner_indices: {left_learner_indices}")
            if not left_learner_indices:
                break
            logger.debug(f"selected_learner_indices: {selected_learner_indices}")
            tmp_reget, next_index, tmp_weighted_ensemble = self.select_next_best(
                left_learner_indices=left_learner_indices,
                selected_learner_indices=selected_learner_indices,
                predictions=predictions,
                labels=labels,
            )
            selected_learner_indices.append(next_index)
            if best_regret is None or tmp_reget < best_regret:
                best_regret = tmp_reget
                best_weighted_ensemble = tmp_weighted_ensemble
                best_selected_learner_indices = copy.deepcopy(selected_learner_indices)
        logger.debug(f"\nbest score: {self._eval_metric_func._optimum-best_regret}")
        logger.debug(f"best_selected_learner_indices: {best_selected_learner_indices}")
        logger.debug(f"best_ensemble_weights: {best_weighted_ensemble.weights_}")

        return best_weighted_ensemble, best_selected_learner_indices

    def update_hyperparameters(self, hyperparameters: Dict):
        if self._hyperparameters and hyperparameters:
            self._hyperparameters.update(hyperparameters)
        elif hyperparameters:
            self._hyperparameters = hyperparameters

        self._hyperparameters = update_ensemble_hyperparameters(
            presets=self._presets,
            provided_hyperparameters=self._hyperparameters,
        )
        # filter out meta-transformer if no local checkpoint path is provided
        if "early_fusion" in self._hyperparameters:
            if self._hyperparameters["early_fusion"]["model.meta_transformer.checkpoint_path"] == "null":
                self._hyperparameters.pop("early_fusion")
                message = (
                    "`early_fusion` will not be used in ensembling because `early_fusion` relies on MetaTransformer, "
                    "but no local MetaTransformer model checkpoint is provided. To use `early_fusion`, "
                    "download its model checkpoints from https://github.com/invictus717/MetaTransformer to local "
                    "and set the checkpoint path as follows:\n"
                    "```python\n"
                    "hyperparameters = {\n"
                    '    "early_fusion": {\n'
                    '        "model.meta_transformer.checkpoint_path": args.meta_transformer_ckpt_path,\n'
                    "    }\n"
                    "}\n"
                    "```\n"
                    "Note that presets `high_quality` (default) and `medium_quality` need the base model, while preset "
                    "`best_quality` requires the large model. Make sure to download the right MetaTransformer version. "
                    "We recommend using the download links under tag `国内下载源` because the corresponding "
                    "downloaded models are not compressed and can be loaded directly.\n"
                )

                logger.warning(message)

    def fit_all(
        self,
        train_data,
        tuning_data,
        hyperparameters,
        column_types,
        holdout_frac,
        time_limit,
        seed,
        standalone,
        clean_ckpts,
    ):
        self._relative_path = True
        self.update_hyperparameters(hyperparameters=hyperparameters)

        learners = []
        assert (
            len(self._hyperparameters) > 1
        ), f"Ensembling requires training more than 1 learners, but got {len(self._hyperparameters)} sets of hyperparameters."
        logger.info(
            f"Will ensemble {len(self._hyperparameters)} models with the following configs:\n {pprint.pformat(self._hyperparameters)}"
        )
        for per_name, per_hparams in self._hyperparameters.items():
            per_learner_path = os.path.join(self._save_path, per_name)
            if not os.path.isdir(per_learner_path):
                logger.info(f"\nfitting learner {per_name}")
                logger.debug(f"hyperparameters: {per_hparams}")
                per_learner = BaseLearner(
                    label=self._label_column,
                    problem_type=self._problem_type,
                    presets=self._presets,
                    eval_metric=self._eval_metric_func,
                    hyperparameters=per_hparams,
                    path=per_learner_path,
                    verbosity=self._verbosity,
                    warn_if_exist=self._warn_if_exist,
                    enable_progress_bar=self._enable_progress_bar,
                    pretrained=self._pretrained,
                    validation_metric=self._validation_metric_name,
                )
                per_learner.fit(
                    train_data=train_data,
                    tuning_data=tuning_data,
                    time_limit=time_limit,
                    column_types=column_types,
                    holdout_frac=holdout_frac,
                    seed=seed,
                    standalone=standalone,
                    clean_ckpts=clean_ckpts,
                )
            learners.append(per_name)

        return learners

    def on_fit_end(
        self,
        training_start: float,
        **kwargs,
    ):
        self._fit_called = True
        training_end = time.time()
        self._total_train_time = training_end - training_start
        logger.info(on_fit_end_message(self._save_path))

    def update_attributes_by_first_learner(self, learners: List):
        # load df preprocessor from the first learner
        if isinstance(learners[0], str):
            first_learner_path = self.get_learner_path(learners[0])
            dir_path, ckpt_path = get_dir_ckpt_paths(path=first_learner_path)
            assert os.path.isdir(dir_path), f"'{dir_path}' must be an existing directory."
            first_learner = BaseLearner(label="dummy_label")
            first_learner = BaseLearner._load_metadata(learner=first_learner, path=dir_path)
        else:
            first_learner = learners[0]

        self._df_preprocessor = first_learner._df_preprocessor
        self._eval_metric_func = first_learner._eval_metric_func
        self._eval_metric_name = first_learner._eval_metric_name
        self._problem_type = first_learner._problem_type

    def fit_ensemble(
        self,
        predictions: Optional[List[np.ndarray]] = None,
        labels: Optional[np.ndarray] = None,
        learners: Optional[List[Union[str, BaseLearner]]] = None,
        train_data: Optional[Union[pd.DataFrame, str]] = None,
        tuning_data: Optional[Union[pd.DataFrame, str]] = None,
        holdout_frac: Optional[float] = None,
        seed: Optional[int] = 0,
    ):
        if not predictions or labels is None:
            self.prepare_train_tuning_data(
                train_data=train_data,
                tuning_data=tuning_data,
                holdout_frac=holdout_frac,
                seed=seed,
            )
            predictions, labels = self.predict_all_for_ensembling(
                learners=learners,
                data=self._tuning_data,
                mode=VAL,
                requires_label=True,
                save=True,
            )

        self.verify_predictions_labels(
            predictions=predictions,
            labels=labels,
            learners=learners,
        )

        if self._ensemble_mode == "sequential":
            weighted_ensemble, selected_learner_indices = self.sequential_ensemble(
                predictions=predictions,
                labels=labels,
            )
        elif self._ensemble_mode == "one_shot":
            weighted_ensemble = self.fit_per_ensemble(
                predictions=predictions,
                labels=labels,
            )
            selected_learner_indices = list(range(len(learners)))
        else:
            raise ValueError(f"Unsupported ensemble_mode: {self._ensemble_mode}")

        predictions = [predictions[j] for j in selected_learner_indices]
        predictions = weighted_ensemble.predict_proba(predictions)

        # for regression, the transform_prediction() is already called in predict_all()
        if self._eval_metric_func.needs_pred and self._problem_type != REGRESSION:
            predictions = self._df_preprocessor.transform_prediction(
                y_pred=predictions,
                inverse_categorical=False,
            )
        metric_data = {
            Y_PRED: predictions,
            Y_TRUE: labels,
        }
        score = compute_score(
            metric_data=metric_data,
            metric=self._eval_metric_func,
        )

        logger.debug(f"\nEnsembling score on validation data: {score}")

        return weighted_ensemble, selected_learner_indices

    def fit(
        self,
        train_data: Union[pd.DataFrame, str],
        presets: Optional[str] = None,
        tuning_data: Optional[Union[pd.DataFrame, str]] = None,
        time_limit: Optional[int] = None,
        save_path: Optional[str] = None,
        hyperparameters: Optional[Union[str, Dict, List[str]]] = None,
        column_types: Optional[Dict] = None,
        holdout_frac: Optional[float] = None,
        teacher_learner: Union[str, BaseLearner] = None,
        seed: Optional[int] = 0,
        standalone: Optional[bool] = True,
        hyperparameter_tune_kwargs: Optional[Dict] = None,
        clean_ckpts: Optional[bool] = True,
        learners: Optional[List[Union[str, BaseLearner]]] = None,
        predictions: Optional[List[np.ndarray]] = None,
        labels: Optional[np.ndarray] = None,
        **kwargs,
    ):
        self.setup_save_path(save_path=save_path)
        training_start = self.on_fit_start(presets=presets)
        if learners is None:
            learners = self.fit_all(
                train_data=train_data,
                tuning_data=tuning_data,
                hyperparameters=hyperparameters,
                column_types=column_types,
                holdout_frac=holdout_frac,
                time_limit=time_limit,
                seed=seed,
                standalone=standalone,
                clean_ckpts=clean_ckpts,
            )
        assert len(learners) > 1, f"Ensembling requires more than 1 learners, but got {len(learners)}."

        self.update_attributes_by_first_learner(learners=learners)
        weighted_ensemble, selected_learner_indices = self.fit_ensemble(
            predictions=predictions,
            labels=labels,
            learners=learners,
            train_data=train_data,
            tuning_data=tuning_data,
            holdout_frac=holdout_frac,
            seed=seed,
        )

        assert len(selected_learner_indices) == len(weighted_ensemble.weights_)
        self._weighted_ensemble = weighted_ensemble
        self._selected_learners = [learners[i] for i in selected_learner_indices]
        self._all_learners = learners
        self._selected_indices = selected_learner_indices

        self.on_fit_end(training_start=training_start)
        self.save(path=self._save_path)

        return self

    def predict(
        self,
        data: Union[pd.DataFrame, dict, list, str],
        predictions: Optional[List[np.ndarray]] = None,
        as_pandas: Optional[bool] = None,
        **kwargs,
    ):
        self.on_predict_start()
        if not predictions:
            predictions = self.predict_all_for_ensembling(
                learners=self._selected_learners,
                data=data,
                mode=TEST,
                requires_label=False,
                save=False,
            )
        else:
            predictions = [predictions[i] for i in self._selected_indices]

        self.verify_predictions_labels(
            predictions=predictions,
            learners=self._selected_learners,
        )
        pred = self._weighted_ensemble.predict_proba(predictions)
        # for regression, the transform_prediction() is already called in predict_all()
        if self._problem_type in [BINARY, MULTICLASS]:
            pred = self._df_preprocessor.transform_prediction(
                y_pred=pred,
                inverse_categorical=True,
            )
        if (as_pandas is None and isinstance(data, pd.DataFrame)) or as_pandas is True:
            pred = self._as_pandas(data=data, to_be_converted=pred)

        return pred

    def predict_proba(
        self,
        data: Union[pd.DataFrame, dict, list],
        predictions: Optional[List[np.ndarray]] = None,
        as_pandas: Optional[bool] = None,
        as_multiclass: Optional[bool] = True,
        **kwargs,
    ):
        self.on_predict_start()
        assert self._problem_type not in [
            REGRESSION,
        ], f"Problem {self._problem_type} has no probability output."

        if not predictions:
            predictions = self.predict_all_for_ensembling(
                learners=self._selected_learners,
                data=data,
                mode=TEST,
                requires_label=False,
                save=False,
            )
        else:
            predictions = [predictions[i] for i in self._selected_indices]

        self.verify_predictions_labels(
            predictions=predictions,
            learners=self._selected_learners,
        )
        prob = self._weighted_ensemble.predict_proba(predictions)
        if as_multiclass and self._problem_type == BINARY:
            prob = np.column_stack((1 - prob, prob))

        if (as_pandas is None and isinstance(data, pd.DataFrame)) or as_pandas is True:
            prob = self._as_pandas(data=data, to_be_converted=prob)

        return prob

    def evaluate(
        self,
        data: Union[pd.DataFrame, dict, list, str],
        predictions: Optional[List[np.ndarray]] = None,
        labels: Optional[np.ndarray] = None,
        save_all: Optional[bool] = True,
        **kwargs,
    ):
        self.on_predict_start()
        if not predictions or labels is None:
            if save_all:
                learners = self._all_learners
            else:
                learners = self._selected_learners
            predictions, labels = self.predict_all_for_ensembling(
                learners=learners,
                data=data,
                mode=TEST,
                requires_label=True,
                save=True,
            )
            if save_all:
                predictions = [predictions[i] for i in self._selected_indices]
        else:
            predictions = [predictions[i] for i in self._selected_indices]

        self.verify_predictions_labels(
            predictions=predictions,
            labels=labels,
            learners=self._selected_learners,
        )
        all_scores = dict()
        for per_predictions, per_learner in zip(predictions, self._selected_learners):
            if not isinstance(per_learner, str):
                per_learner = per_learner.path
            metric_data = {
                Y_PRED: per_predictions,
                Y_TRUE: labels,
            }
            all_scores[per_learner] = compute_score(
                metric_data=metric_data,
                metric=self._eval_metric_func,
            )

        predictions = self._weighted_ensemble.predict_proba(predictions)
        # for regression, the transform_prediction() is already called in predict_all()
        if self._eval_metric_func.needs_pred and self._problem_type != REGRESSION:
            predictions = self._df_preprocessor.transform_prediction(
                y_pred=predictions,
                inverse_categorical=False,
            )
        metric_data = {
            Y_PRED: predictions,
            Y_TRUE: labels,
        }
        all_scores["ensemble"] = compute_score(
            metric_data=metric_data,
            metric=self._eval_metric_func,
        )

        return all_scores

    def extract_embedding(
        self,
        data: Union[pd.DataFrame, dict, list],
        return_masks: Optional[bool] = False,
        as_tensor: Optional[bool] = False,
        as_pandas: Optional[bool] = False,
        realtime: Optional[bool] = False,
        **kwargs,
    ):
        raise ValueError(f"EnsembleLearner doesn't support extracting embedding yet.")

    def save(
        self,
        path: str,
        **kwargs,
    ):
        selected_learner_names = [self.get_learner_name(per_learner) for per_learner in self._selected_learners]
        all_learner_names = [self.get_learner_name(per_learner) for per_learner in self._all_learners]

        os.makedirs(path, exist_ok=True)
        with open(os.path.join(path, f"assets.json"), "w") as fp:
            json.dump(
                {
                    "learner_class": self.__class__.__name__,
                    "ensemble_size": self._ensemble_size,
                    "ensemble_mode": self._ensemble_mode,
                    "selected_learners": selected_learner_names,
                    "all_learners": all_learner_names,
                    "selected_indices": self._selected_indices,
                    "ensemble_weights": self._weighted_ensemble.weights_.tolist(),
                    "save_path": path,
                    "relative_path": True,
                    "fit_called": self._fit_called,
                    "version": ag_version.__version__,
                    "hyperparameters": self._hyperparameters,
                },
                fp,
                ensure_ascii=True,
            )

        with open(os.path.join(path, "ensemble.pkl"), "wb") as fp:
            pickle.dump(self._weighted_ensemble, fp)

        # save each learner
        for per_learner in self._all_learners:
            per_learner_name = self.get_learner_name(per_learner)
            if isinstance(per_learner, str):
                per_learner_path = self.get_learner_path(per_learner)
                per_learner = BaseLearner.load(per_learner_path)

            per_learner_save_path = os.path.join(path, per_learner_name)
            per_learner.save(per_learner_save_path)

        return

    @classmethod
    def load(
        cls,
        path: str,
        **kwargs,
    ):
        dir_path, ckpt_path = get_dir_ckpt_paths(path=path)
        assert os.path.isdir(dir_path), f"'{dir_path}' must be an existing directory."
        with open(os.path.join(dir_path, "assets.json"), "r") as fp:
            assets = json.load(fp)

        learner = cls(
            hyperparameters=assets["hyperparameters"],
        )
        learner._ensemble_size = assets["ensemble_size"]
        learner._ensemble_mode = assets["ensemble_mode"]
        learner._selected_learners = assets["selected_learners"]
        learner._all_learners = assets["all_learners"]
        learner._selected_indices = assets["selected_indices"]
        learner._save_path = path  # in case the original exp dir is copied to somewhere else
        learner._relative_path = assets["relative_path"]
        learner._fit_called = assets["fit_called"]

        with open(os.path.join(path, "ensemble.pkl"), "rb") as fp:
            learner._weighted_ensemble = pickle.load(fp)  # nosec B301

        learner.update_attributes_by_first_learner(learners=learner._selected_learners)

        return learner



import base64
import json
import logging
import os
import re

import numpy as np
import pandas as pd
import torch
from scipy.special import expit, softmax

logger = logging.getLogger(__name__)


def logits_to_prob(logits: np.ndarray):
    """
    Convert logits to probabilities.

    Parameters
    ----------
    logits
        The logits output of a classification head.

    Returns
    -------
    Probabilities.
    """
    if logits.ndim == 1:
        return expit(logits)
    elif logits.ndim == 2:
        return softmax(logits, axis=1)
    else:
        raise ValueError(f"Unsupported logit dim: {logits.ndim}.")


def tensor_to_ndarray(tensor: torch.Tensor):
    """
    Convert Pytorch tensor to numpy array.

    Parameters
    ----------
    tensor
        A Pytorch tensor.

    Returns
    -------
    A ndarray.
    """
    return tensor.detach().cpu().float().numpy()


def path_expander(path, base_folder):
    path_l = path.split(";")
    return ";".join([os.path.abspath(os.path.join(base_folder, path)) for path in path_l])


def _read_byte(file):
    with open(file, "rb") as image:
        f = image.read()
        b = bytearray(f)
    return b


def path_to_bytearray_expander(path, base_folder):
    path_l = path.split(";")
    return [_read_byte(os.path.abspath(os.path.join(base_folder, path))) for path in path_l]


def _read_base64str(file):
    with open(file, "rb") as image:
        f = image.read()
        image_base64 = base64.b64encode(f)
        image_base64_str = image_base64.decode("utf-8")
    return image_base64_str


def path_to_base64str_expander(path, base_folder):
    path_l = path.split(";")
    return [_read_base64str(os.path.abspath(os.path.join(base_folder, path))) for path in path_l]


def shopee_dataset(
    download_dir: str,
    is_bytearray=False,
    is_base64str=False,
):
    """
    Download Shopee dataset for demo.

    Parameters
    ----------
    download_dir
        Path to save the dataset locally.

    Returns
    -------
    train and test set of Shopee dataset in pandas DataFrame format.
    """
    zip_file = "https://automl-mm-bench.s3.amazonaws.com/vision_datasets/shopee.zip"
    from autogluon.core.utils.loaders import load_zip

    load_zip.unzip(zip_file, unzip_dir=download_dir)

    dataset_path = os.path.join(download_dir, "shopee")

    train_data = pd.read_csv(f"{dataset_path}/train.csv")
    test_data = pd.read_csv(f"{dataset_path}/test.csv")

    expander = (
        path_to_bytearray_expander if is_bytearray else (path_to_base64str_expander if is_base64str else path_expander)
    )
    train_data["image"] = train_data["image"].apply(lambda ele: expander(ele, base_folder=dataset_path))
    test_data["image"] = test_data["image"].apply(lambda ele: expander(ele, base_folder=dataset_path))
    return train_data, test_data


def merge_spans(sent, pred, for_visualizer=False):
    """Merge subsequent predictions."""
    if isinstance(pred, str):
        # For string values, we assume that it is json-encoded string of the sentences.
        try:
            pred = json.loads(pred)
        except Exception as exp:
            raise RuntimeError(
                f"The received entity annotations is {pred}, "
                f"which can not be encoded with the json format. "
                f"Check your input again, or running `json.loads(pred)` to verify your data."
            )
    spans = {}
    last_start = -1
    last_end = -1
    last_label = ""
    for entity in pred:
        entity_group = entity["entity_group"]
        start = entity["start"]
        end = entity["end"]
        if (
            last_start >= 0
            and not for_visualizer
            and (not re.match("B-", entity_group, re.IGNORECASE))
            and (
                (re.match("I-", entity_group, re.IGNORECASE) and last_label[2:] == entity_group[2:])
                or last_label == entity_group
            )
            and (sent[last_end:start].isspace() or (last_end == start))
        ):
            last_end = end
        else:
            last_start = start
            last_end = end
            last_label = entity_group

        if re.match("B-", last_label, re.IGNORECASE) or re.match("I-", last_label, re.IGNORECASE):
            spans.update({last_start: (last_end, last_label[2:])})
        else:
            spans.update({last_start: (last_end, last_label)})
    return spans


def merge_bio_format(data, preds):
    """Merge predictions with BIO format during prediction."""
    results = []
    for sent, pred in zip(data, preds):
        results.append(
            [
                {"entity_group": value[-1], "start": key, "end": value[0]}
                for key, value in merge_spans(sent, pred).items()
            ]
        )
    return results



# model settings
model = dict(
    type="FasterRCNN",
    backbone=dict(
        type="ResNet",
        depth=50,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        norm_eval=True,
        style="pytorch",
        init_cfg=dict(type="Pretrained", checkpoint="torchvision://resnet50"),
    ),
    neck=dict(type="FPN", in_channels=[256, 512, 1024, 2048], out_channels=256, num_outs=5),
    rpn_head=dict(
        type="RPNHead",
        in_channels=256,
        feat_channels=256,
        anchor_generator=dict(type="AnchorGenerator", scales=[8], ratios=[0.5, 1.0, 2.0], strides=[4, 8, 16, 32, 64]),
        bbox_coder=dict(
            type="DeltaXYWHBBoxCoder", target_means=[0.0, 0.0, 0.0, 0.0], target_stds=[1.0, 1.0, 1.0, 1.0]
        ),
        loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=True, loss_weight=1.0),
        loss_bbox=dict(type="L1Loss", loss_weight=1.0),
    ),
    roi_head=dict(
        type="StandardRoIHead",
        bbox_roi_extractor=dict(
            type="SingleRoIExtractor",
            roi_layer=dict(type="RoIAlign", output_size=7, sampling_ratio=0),
            out_channels=256,
            featmap_strides=[4, 8, 16, 32],
        ),
        bbox_head=dict(
            type="Shared2FCBBoxHead",
            in_channels=256,
            fc_out_channels=1024,
            roi_feat_size=7,
            num_classes=80,
            bbox_coder=dict(
                type="DeltaXYWHBBoxCoder", target_means=[0.0, 0.0, 0.0, 0.0], target_stds=[0.1, 0.1, 0.2, 0.2]
            ),
            reg_class_agnostic=False,
            loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0),
            loss_bbox=dict(type="L1Loss", loss_weight=1.0),
        ),
    ),
    # model training and testing settings
    train_cfg=dict(
        rpn=dict(
            assigner=dict(
                type="MaxIoUAssigner",
                pos_iou_thr=0.7,
                neg_iou_thr=0.3,
                min_pos_iou=0.3,
                match_low_quality=True,
                ignore_iof_thr=-1,
            ),
            sampler=dict(type="RandomSampler", num=256, pos_fraction=0.5, neg_pos_ub=-1, add_gt_as_proposals=False),
            allowed_border=-1,
            pos_weight=-1,
            debug=False,
        ),
        rpn_proposal=dict(nms_pre=2000, max_per_img=1000, nms=dict(type="nms", iou_threshold=0.7), min_bbox_size=0),
        rcnn=dict(
            assigner=dict(
                type="MaxIoUAssigner",
                pos_iou_thr=0.5,
                neg_iou_thr=0.5,
                min_pos_iou=0.5,
                match_low_quality=False,
                ignore_iof_thr=-1,
            ),
            sampler=dict(type="RandomSampler", num=512, pos_fraction=0.25, neg_pos_ub=-1, add_gt_as_proposals=True),
            pos_weight=-1,
            debug=False,
        ),
    ),
    test_cfg=dict(
        rpn=dict(nms_pre=1000, max_per_img=1000, nms=dict(type="nms", iou_threshold=0.7), min_bbox_size=0),
        rcnn=dict(score_thr=0.05, nms=dict(type="nms", iou_threshold=0.5), max_per_img=100),
        # soft-nms is also supported for rcnn testing
        # e.g., nms=dict(type='soft_nms', iou_threshold=0.5, min_score=0.05)
    ),
)



_base_ = ["../faster_rcnn/faster_rcnn_r50_fpn.py", "./voc0712.py", "../default_runtime.py"]
model = dict(roi_head=dict(bbox_head=dict(num_classes=20)))
# optimizer
optimizer = dict(type="SGD", lr=0.01, momentum=0.9, weight_decay=0.0001)
optimizer_config = dict(grad_clip=None)
# learning policy
# actual epoch = 3 * 3 = 9
lr_config = dict(policy="step", step=[3])
# runtime settings
runner = dict(type="EpochBasedRunner", max_epochs=4)  # actual epoch = 4 * 3 = 12



import logging
from typing import Callable, Dict, List, Optional, Union

import torch
import torchmetrics
from torch import nn
from torch.nn.modules.loss import _Loss

from ..constants import FUSION_NER, LOGITS, NER_TEXT, WEIGHT
from ..data.mixup import MixupModule
from .lit_module import LitModule

logger = logging.getLogger(__name__)


class NerLitModule(LitModule):
    """
    Control the loops for training, evaluation, and prediction of Named Entity Recognition. This module is independent of
    the model definition. This class inherits from Lightning's LightningModule:
    https://lightning.ai/docs/pytorch/stable/common/lightning_module.html
    """

    def __init__(
        self,
        model: nn.Module,
        optim_type: Optional[str] = None,
        lr_choice: Optional[str] = None,
        lr_schedule: Optional[str] = None,
        lr: Optional[float] = None,
        lr_decay: Optional[float] = None,
        end_lr: Optional[Union[float, int]] = None,
        lr_mult: Optional[Union[float, int]] = None,
        weight_decay: Optional[float] = None,
        warmup_steps: Optional[int] = None,
        loss_func: Optional[_Loss] = None,
        validation_metric: Optional[torchmetrics.Metric] = None,
        validation_metric_name: Optional[str] = None,
        custom_metric_func: Callable = None,
        test_metric: Optional[torchmetrics.Metric] = None,
        peft: Optional[str] = None,
        trainable_param_names: Optional[List] = None,
        mixup_fn: Optional[MixupModule] = None,
        mixup_off_epoch: Optional[int] = 0,
        model_postprocess_fn: Callable = None,
        skip_final_val: Optional[bool] = False,
        track_grad_norm: Optional[Union[int, str]] = -1,
    ):
        """
        Parameters
        ----------
        model
            A Pytorch model
        optim_type
            Optimizer type. We now support:
            - adamw
            - adam
            - sgd
        lr_choice
            How to set each layer's learning rate. If not specified, the default is a single
            learnng rate for all layers. Otherwise, we now support two choices:
            - two_stages
                The layers in the pretrained models have a small learning rate (lr * lr_mult),
                while the newly added head layers use the provided learning rate.
            - layerwise_decay
                The layers have decreasing learning rate from the output end to the input end.
                The intuition is that later layers are more task-related, hence larger learning rates.
        lr_schedule
            Learning rate schedule. We now support:
            - cosine_decay
                Linear warmup followed by cosine decay
            - polynomial_decay
                Linear warmup followed by polynomial decay
        lr
            Learning rate.
        lr_decay
            The learning rate decay factor (0, 1). It is used only when lr_choice is "layerwise_decay".
        end_lr
            The final learning rate after decay.
        lr_mult
            The learning rate multiplier (0, 1). It is used only when lr_choice is "two_stages".
        weight_decay
            The weight decay to regularize layer weights' l2 norm.
        warmup_steps
            How many steps to warmup learning rate. If a float (0, 1), it would represent the
            percentage of steps over all the training steps. The actual number is calculated as
            "int(warmup_steps * max_steps)". If an integer, it would be the exact step number.
        loss_func
            A Pytorch loss module, e.g., nn.CrossEntropyLoss().
        validation_metric
            A torchmetrics module used in the validation stage, e.g., torchmetrics.Accuracy().
        validation_metric_name
            Name of validation metric in case that validation_metric is a aggregation metric,
            e.g., torchmetrics.MeanMetric, whose name can't reflect the real metric name.
        custom_metric_func
            A customized metric function in case that torchmetrics doesn't have the metric.
            It is generally used together with torchmetrics' aggregators, e.g., torchmetrics.MeanMetric.
            Refer to https://github.com/PyTorchLightning/metrics/blob/master/torchmetrics/aggregation.py
        test_metric
            A torchmetrics module used in the test stage, e.g., torchmetrics.Accuracy().
        peft
            Whether to use efficient finetuning strategies. This will be helpful for fast finetuning of large backbones.
            We support options such as:

            - bit_fit (only finetune the bias terms)
            - norm_fit (only finetune the weights in norm layers / bias layer)
            - lora, lora_bias, lora_norm (only finetunes decomposition matrices inserted into model, in combination with either bit_fit or norm_fit)
            - ia3, ia3_bias, ia3_norm (adds vector that scales activations by learned vectors, in combination with either bit_fit or norm_fit)
            - None (do not use efficient finetuning strategies)
        track_grad_norm
            Track the p-norm of gradients during training. May be set to ‘inf’ infinity-norm.
            If using Automatic Mixed Precision (AMP), the gradients will be unscaled before logging them.

        """
        super().__init__(
            model=model,
            optim_type=optim_type,
            lr_choice=lr_choice,
            lr_schedule=lr_schedule,
            lr=lr,
            lr_decay=lr_decay,
            end_lr=end_lr,
            lr_mult=lr_mult,
            weight_decay=weight_decay,
            warmup_steps=warmup_steps,
            loss_func=loss_func,
            validation_metric=validation_metric,
            validation_metric_name=validation_metric_name,
            custom_metric_func=custom_metric_func,
            test_metric=test_metric,
            peft=peft,
            trainable_param_names=trainable_param_names,
            mixup_fn=mixup_fn,
            mixup_off_epoch=mixup_off_epoch,
            model_postprocess_fn=model_postprocess_fn,
            skip_final_val=skip_final_val,
            track_grad_norm=track_grad_norm,
        )

    def _compute_loss(
        self,
        output: Dict,
        label: torch.Tensor,
    ):
        loss = 0
        for prefix, per_output in output.items():
            if prefix == NER_TEXT or prefix == FUSION_NER:
                weight = per_output[WEIGHT] if WEIGHT in per_output else 1
                active_loss = label.view(-1) != 0
                active_logits = per_output[LOGITS].view(-1, self.model.num_classes)[active_loss]
                active_labels = label.view(-1)[active_loss]
                loss += (
                    self.loss_func(
                        input=active_logits,
                        target=active_labels,
                    )
                    * weight
                )
        return loss

    def validation_step(self, batch, batch_idx):
        """
        Per validation step. This function is registered by LightningModule.
        Refer to https://lightning.ai/docs/pytorch/stable/common/lightning_module.html#validation

        Parameters
        ----------
        batch
            A dictionary containing the mini-batch data, including both input data and
            ground-truth labels. The mini-batch data are passed to each individual model,
            which indexes its required input data by keys with its model prefix. The
            ground-truth labels are used here to compute the validation loss and metric.
            The validation metric is used for top k model selection and early stopping.
        batch_idx
            Index of mini-batch.
        """
        output, loss = self._shared_step(batch)
        if self.model_postprocess_fn:
            output = self.model_postprocess_fn(output)

        # By default, on_step=False and on_epoch=True
        self.log("val_loss", loss)
        label = batch[self.model.label_key]
        logits = output[self.model.prefix][LOGITS]
        active_loss = label.view(-1) != 0
        active_logits = logits.view(-1, self.model.num_classes)[active_loss]
        active_labels = label.view(-1)[active_loss]
        self._compute_metric_score(
            metric=self.validation_metric,
            custom_metric_func=self.custom_metric_func,
            logits=active_logits,
            label=active_labels,
        )
        self.log(
            self.validation_metric_name,
            self.validation_metric,
            on_step=False,
            on_epoch=True,
        )



# optimizer
optimizer = dict(type="SGD", lr=0.02, momentum=0.9, weight_decay=0.0001)
optimizer_config = dict(grad_clip=None)
# learning policy
lr_config = dict(policy="step", warmup="linear", warmup_iters=500, warmup_ratio=0.001, step=[8, 11])
runner = dict(type="EpochBasedRunner", max_epochs=12)
